// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Dump current IPv6 state information.
//

#include <winsock2.h>
#include <ws2ip6.h>
#include <stdio.h>
#include <stdlib.h>


HANDLE Handle;

void QueryInterface(int argc, char *argv[]);
void QueryNeighborCache(int argc, char *argv[]);
void QueryRouteCache(int argc, char *argv[]);
void QueryPrefixList(int argc, char *argv[]);
void QueryRouterList(int argc, char *argv[]);

void
usage(void)
{
    printf("usage: ipv6 if [index]\n");
    printf("       ipv6 nc [index [address]]\n");
    printf("       ipv6 rc [index address]\n");
    printf("       ipv6 pl [index prefix]\n");
    printf("       ipv6 rl [index address]\n");
    exit(1);
}

int _CRTAPI1
main(int argc, char **argv)
{
    WSADATA wsaData;

    //
    // We initialize Winsock only to have access
    // to WSAStringToAddress and WSAAddressToString.
    //
    if (WSAStartup(MAKEWORD(2, 0), &wsaData)) {
        printf("WSAStartup failed\n");
        exit(1);
    }

    if (argc < 2) {
        usage();
    }

    Handle = CreateFileW(WIN_IPV6_DEVICE_NAME,
                         0,      // access mode
                         FILE_SHARE_READ | FILE_SHARE_WRITE,
                         NULL,   // security attributes
                         OPEN_EXISTING,
                         0,      // flags & attributes
                         NULL);  // template file
    if (Handle == INVALID_HANDLE_VALUE) {
        printf("Could not access IPv6 protocol stack\n");
        exit(1);
    }

    if (!strcmp(argv[1], "if")) {
        QueryInterface(argc - 2, argv + 2);
    }
    else if (!strcmp(argv[1], "nc")) {
        QueryNeighborCache(argc - 2, argv + 2);
    }
    else if (!strcmp(argv[1], "rc")) {
        QueryRouteCache(argc - 2, argv + 2);
    }
    else if (!strcmp(argv[1], "pl")) {
        QueryPrefixList(argc - 2, argv + 2);
    }
    else if (!strcmp(argv[1], "rl")) {
        QueryRouterList(argc - 2, argv + 2);
    }
    else {
        usage();
    }

    return 0;
}

#define MAX_LINK_LEVEL_ADDRESS_LENGTH   64

int
GetAddress(char *astr, IPv6Addr *address)
{
    struct sockaddr_in6 sin6;
    int addrlen = sizeof sin6;

    sin6.sin6_family = AF_INET6; // shouldn't be required but is

    if ((WSAStringToAddress(astr, AF_INET6, NULL,
                           (struct sockaddr *)&sin6, &addrlen)
                    == SOCKET_ERROR) ||
        (sin6.sin6_port != 0))
        return FALSE;

    // The user gave us a numeric IPv6 address.

    memcpy(address, &sin6.sin6_addr, sizeof *address);
    return TRUE;
}

int
GetPrefix(char *astr, IPv6Addr *prefix, uint *prefixlen)
{
    struct sockaddr_in6 sin6;
    int addrlen = sizeof sin6;
    char *slash;
    int length;

    slash = strchr(astr, '/');
    if (slash == NULL)
        return FALSE;
    *slash = '\0';

    length = atoi(slash+1);
    if ((length <= 0) || (length > 128))
        return FALSE;

    sin6.sin6_family = AF_INET6; // shouldn't be required but is

    if ((WSAStringToAddress(astr, AF_INET6, NULL,
                           (struct sockaddr *)&sin6, &addrlen)
                    == SOCKET_ERROR) ||
        (sin6.sin6_port != 0))
        return FALSE;

    // The user gave us a numeric IPv6 address.

    memcpy(prefix, &sin6.sin6_addr, sizeof *prefix);
    *prefixlen = length;
    return TRUE;
}

char *
FormatIPv6Address(IPv6Addr *Address)
{
    static char buffer[128];
    DWORD buflen = sizeof buffer;
    struct sockaddr_in6 sin6;

    sin6.sin6_family = AF_INET6;
    sin6.sin6_port = 0;
    sin6.sin6_flowinfo = 0;
    memcpy(&sin6.sin6_addr, Address, sizeof *Address);

    if (WSAAddressToString((struct sockaddr *) &sin6,
                           sizeof sin6,
                           NULL,       // LPWSAPROTOCOL_INFO
                           buffer,
                           &buflen) == SOCKET_ERROR)
        strcpy(buffer, "<invalid>");

    return buffer;
}

char *
FormatIPv4Address(struct in_addr *Address)
{
    static char buffer[128];
    DWORD buflen = sizeof buffer;
    struct sockaddr_in sin;

    sin.sin_family = AF_INET;
    sin.sin_port = 0;
    sin.sin_addr = *Address;

    if (WSAAddressToString((struct sockaddr *) &sin,
                           sizeof sin,
                           NULL,       // LPWSAPROTOCOL_INFO
                           buffer,
                           &buflen) == SOCKET_ERROR)
        strcpy(buffer, "<invalid>");

    return buffer;
}

char *
FormatLinkLevelAddress(uint length, uchar *addr)
{
    static char buffer[128];

    switch (length) {
    case 6: {
        int i, digit;
        char *s = buffer;

        for (i = 0; i < 6; i++) {
            if (i != 0)
                *s++ = '-';

            digit = addr[i] >> 4;
            if (digit < 10)
                *s++ = digit + '0';
            else
                *s++ = digit - 10 + 'a';

            digit = addr[i] & 0xf;
            if (digit < 10)
                *s++ = digit + '0';
            else
                *s++ = digit - 10 + 'a';
        }
        *s = '\0';
        break;
    }

    case 4:
        //
        // IPv4 address (6-over-4 link)
        //
        strcpy(buffer, FormatIPv4Address((struct in_addr *)addr));
        break;

    case 0:
        //
        // Null or loop-back address
        //
        buffer[0] = '\0';
        break;

    default:
        printf("unrecognized link-level address format\n");
        exit(1);
    }

    return buffer;
}

void
ForEachAddress(IPV6_INFO_INTERFACE *IF,
               void (*func)(IPV6_INFO_ADDRESS *))
{
    IPV6_QUERY_ADDRESS Query, NextQuery;
    IPV6_INFO_ADDRESS NTE;
    uint BytesReturned;

    NextQuery.Index = IF->Query.Index;
    NextQuery.Address = in6addr_any;

    for (;;) {
        Query = NextQuery;

        if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_ADDRESS,
                             &Query, sizeof Query,
                             &NTE, sizeof NTE, &BytesReturned,
                             NULL)) {
            printf("bad address %s\n", FormatIPv6Address(&Query.Address));
            exit(1);
        }

        NextQuery = NTE.Query;

        if (!IN6_ADDR_EQUAL(&Query.Address, &in6addr_any)) {

            NTE.Query = Query;
            (*func)(&NTE);
        }

        if (IN6_ADDR_EQUAL(&NextQuery.Address, &in6addr_any))
            break;
    }
}

char *
FormatDADState(uint DADState)
{
    switch (DADState) {
    case 0:
        return "tentative";
    case 1:
        return "preferred";
    case 2:
        return "deprecated";
    case 3:
        return "invalid";
    case 4:
        return "duplicate";
    }
    return "<bad state>";
}

void
PrintAddress(IPV6_INFO_ADDRESS *NTE)
{
    printf("    %s address %s/%u, ",
           FormatDADState(NTE->DADState),
           FormatIPv6Address(&NTE->Query.Address), NTE->PrefixLength);

    if (NTE->ValidLifetime == 0xffffffff)
        printf("infinite/");
    else
        printf("%us/", NTE->ValidLifetime);

    if (NTE->PreferredLifetime == 0xffffffff)
        printf("infinite\n");
    else
        printf("%us\n", NTE->PreferredLifetime);
}

IPV6_INFO_INTERFACE *
GetInterface(uint Index)
{
    IPV6_QUERY_INTERFACE Query;
    IPV6_INFO_INTERFACE *IF;
    uint InfoSize, BytesReturned;

    Query.Index = Index;

    InfoSize = sizeof *IF + MAX_LINK_LEVEL_ADDRESS_LENGTH;
    IF = (IPV6_INFO_INTERFACE *) malloc(InfoSize);
    if (IF == NULL) {
        printf("malloc failed\n");
        exit(1);
    }
    
    if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_INTERFACE,
                         &Query, sizeof Query,
                         IF, InfoSize, &BytesReturned,
                         NULL)) {
        printf("bad index %u\n", Query.Index);
        exit(1);
    }

    if (BytesReturned != sizeof *IF + IF->LinkLevelAddressLength) {
        printf("inconsistent link-level address length\n");
        exit(1);
    }

    IF->Query = Query;
    return IF;
}

void
ForEachInterface(void (*func)(IPV6_INFO_INTERFACE *))
{
    IPV6_QUERY_INTERFACE Query, NextQuery;
    IPV6_INFO_INTERFACE *IF;
    uint InfoSize, BytesReturned;

    InfoSize = sizeof *IF + MAX_LINK_LEVEL_ADDRESS_LENGTH;
    IF = (IPV6_INFO_INTERFACE *) malloc(InfoSize);
    if (IF == NULL) {
        printf("malloc failed\n");
        exit(1);
    }

    NextQuery.Index = 0;

    for (;;) {
        Query = NextQuery;

        if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_INTERFACE,
                             &Query, sizeof Query,
                             IF, InfoSize, &BytesReturned,
                             NULL)) {
            printf("bad index %u\n", Query.Index);
            exit(1);
        }

        NextQuery = IF->Query;

        if (Query.Index != 0) {

            if (BytesReturned != sizeof *IF + IF->LinkLevelAddressLength) {
                printf("inconsistent link-level address length\n");
                exit(1);
            }

            IF->Query = Query;
            (*func)(IF);
        }

        if (NextQuery.Index == 0)
            break;
    }

    free(IF);
}

void
PrintInterface(IPV6_INFO_INTERFACE *IF)
{
    printf("Interface %u:\n", IF->Query.Index);
    printf("  link-level address: %s\n",
           FormatLinkLevelAddress(IF->LinkLevelAddressLength,
                                  (uchar *)(IF + 1)));

    ForEachAddress(IF, PrintAddress);

    printf("  link MTU %u (true link MTU %u)\n",
           IF->LinkMTU, IF->TrueLinkMTU);
    printf("  current hop limit %u\n", IF->CurHopLimit);
    printf("  reachable time %ums (base %ums)\n",
           IF->ReachableTime, IF->BaseReachableTime);
    printf("  retransmission interval %us\n", IF->RetransTimer);
    printf("  DAD transmits %u\n", IF->DupAddrDetectTransmits);
}

IPV6_INFO_NEIGHBOR_CACHE *
GetNeighborCacheEntry(IPV6_INFO_INTERFACE *IF, IPv6Addr *Address)
{
    IPV6_QUERY_NEIGHBOR_CACHE Query;
    IPV6_INFO_NEIGHBOR_CACHE *NCE;
    uint InfoSize, BytesReturned;

    InfoSize = sizeof *NCE + MAX_LINK_LEVEL_ADDRESS_LENGTH;
    NCE = (IPV6_INFO_NEIGHBOR_CACHE *) malloc(InfoSize);
    if (NCE == NULL) {
        printf("malloc failed\n");
        exit(1);
    }

    Query.Index = IF->Query.Index;
    Query.Address = *Address;

    if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_NEIGHBOR_CACHE,
                         &Query, sizeof Query,
                         NCE, InfoSize, &BytesReturned,
                         NULL)) {
        printf("bad address %s\n", FormatIPv6Address(&Query.Address));
        exit(1);
    }

    if (BytesReturned != sizeof *NCE + NCE->LinkLevelAddressLength) {
        printf("inconsistent link-level address length\n");
        exit(1);
    }

    NCE->Query = Query;
    return NCE;
}

void
ForEachNeighborCacheEntry(IPV6_INFO_INTERFACE *IF,
                          void (*func)(IPV6_INFO_NEIGHBOR_CACHE *))
{
    IPV6_QUERY_NEIGHBOR_CACHE Query, NextQuery;
    IPV6_INFO_NEIGHBOR_CACHE *NCE;
    uint InfoSize, BytesReturned;

    InfoSize = sizeof *NCE + MAX_LINK_LEVEL_ADDRESS_LENGTH;
    NCE = (IPV6_INFO_NEIGHBOR_CACHE *) malloc(InfoSize);
    if (NCE == NULL) {
        printf("malloc failed\n");
        exit(1);
    }

    NextQuery.Index = IF->Query.Index;
    NextQuery.Address = in6addr_any;

    for (;;) {
        Query = NextQuery;

        if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_NEIGHBOR_CACHE,
                             &Query, sizeof Query,
                             NCE, InfoSize, &BytesReturned,
                             NULL)) {
            printf("bad address %s\n", FormatIPv6Address(&Query.Address));
            exit(1);
        }

        NextQuery = NCE->Query;

        if (!IN6_ADDR_EQUAL(&Query.Address, &in6addr_any)) {

            if (BytesReturned != sizeof *NCE + NCE->LinkLevelAddressLength) {
                printf("inconsistent link-level address length\n");
                exit(1);
            }

            NCE->Query = Query;
            (*func)(NCE);
        }

        if (IN6_ADDR_EQUAL(&NextQuery.Address, &in6addr_any))
            break;
    }

    free(NCE);
}

void
PrintNeighborCacheEntry(IPV6_INFO_NEIGHBOR_CACHE *NCE)
{
    printf("%u: %18s", NCE->Query.Index,
           FormatIPv6Address(&NCE->Query.Address));

    if (NCE->NDState != 0)
        printf(" %-17s", FormatLinkLevelAddress(NCE->LinkLevelAddressLength,
                                                (uchar *)(NCE + 1)));
    else
        printf(" %-17s", "");

    switch (NCE->NDState) {
    case 0:
        printf(" incomplete");
        break;
    case 1:
        printf(" probe");
        break;
    case 2:
        printf(" delay");
        break;
    case 3:
        printf(" stale");
        break;
    case 4:
        printf(" reachable (%us)",
               (GetTickCount() - NCE->LastReachable) / 1000);
        break;
    case 5:
        printf(" permanent");
        break;
    }

    if (NCE->IsRouter)
        printf(" (router)");

    printf("\n");
}

void
QueryInterface(int argc, char *argv[])
{
    uint Index;
    IPV6_INFO_INTERFACE *IF;

    if (argc == 0) {
        ForEachInterface(PrintInterface);
    }
    else if (argc == 1) {

        if ((Index = atoi(argv[0])) == 0)
            usage();

        IF = GetInterface(Index);
        PrintInterface(IF);
        free(IF);
    }
    else {
        usage();
    }
}

void
PrintNeighborCache(IPV6_INFO_INTERFACE *IF)
{
    ForEachNeighborCacheEntry(IF, PrintNeighborCacheEntry);
}

void
QueryNeighborCache(int argc, char *argv[])
{
    uint Index;
    IPV6_INFO_INTERFACE *IF;
    IPv6Addr Address;
    IPV6_INFO_NEIGHBOR_CACHE *NCE;

    if (argc == 0) {
        ForEachInterface(PrintNeighborCache);
    }
    else if (argc == 1) {

        if ((Index = atoi(argv[0])) == 0)
            usage();

        IF = GetInterface(Index);
        PrintNeighborCache(IF);
        free(IF);
    }
    else if (argc == 2) {

        if ((Index = atoi(argv[0])) == 0)
            usage();

        if (! GetAddress(argv[1], &Address))
            usage();

        IF = GetInterface(Index);
        NCE = GetNeighborCacheEntry(IF, &Address);
        PrintNeighborCacheEntry(NCE);
        free(NCE);
        free(IF);
    }
    else {
        usage();
    }
}

IPV6_INFO_ROUTE_CACHE *
GetRouteCacheEntry(uint Index, IPv6Addr *Address)
{
    IPV6_QUERY_ROUTE_CACHE Query;
    IPV6_INFO_ROUTE_CACHE *RCE;
    uint BytesReturned;

    Query.Index = Index;
    Query.Address = *Address;

    RCE = (IPV6_INFO_ROUTE_CACHE *) malloc(sizeof *RCE);
    if (RCE == NULL) {
        printf("malloc failed\n");
        exit(1);
    }

    if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_ROUTE_CACHE,
                         &Query, sizeof Query,
                         RCE, sizeof *RCE, &BytesReturned,
                         NULL)) {
        printf("bad index or address\n");
        exit(1);
    }

    RCE->Query = Query;
    return RCE;
}

void
ForEachRoute(void (*func)(IPV6_INFO_ROUTE_CACHE *))
{
    IPV6_QUERY_ROUTE_CACHE Query, NextQuery;
    IPV6_INFO_ROUTE_CACHE RCE;
    uint BytesReturned;

    NextQuery.Index = 0;

    for (;;) {
        Query = NextQuery;

        if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_ROUTE_CACHE,
                             &Query, sizeof Query,
                             &RCE, sizeof RCE, &BytesReturned,
                             NULL)) {
            printf("bad index %u\n", Query.Index);
            exit(1);
        }

        NextQuery = RCE.Query;

        if (Query.Index != 0) {

            RCE.Query = Query;
            (*func)(&RCE);
        }

        if (NextQuery.Index == 0)
            break;
    }
}

void
PrintRouteCacheEntry(IPV6_INFO_ROUTE_CACHE *RCE)
{
    printf("%u: %18s -> ", RCE->Query.Index,
           FormatIPv6Address(&RCE->SourceAddress));
    printf("%-18s : ", FormatIPv6Address(&RCE->Query.Address));
    printf("%-16s PMTU %u",
           FormatIPv6Address(&RCE->NextHopAddress),
           RCE->PathMTU);

    switch (RCE->State) {
    case 0:
        printf(" (perm)");
        break;
    case 1:
        break;
    case 2:
        printf(" (redirect)");
        break;
    }

    printf("\n");
}

void
QueryRouteCache(int argc, char *argv[])
{
    uint Index;
    IPv6Addr Address;
    IPV6_INFO_ROUTE_CACHE *RCE;

    if (argc == 0) {
        ForEachRoute(PrintRouteCacheEntry);
    }
    else if (argc == 2) {

        if ((Index = atoi(argv[0])) == 0)
            usage();

        if (! GetAddress(argv[1], &Address))
            usage();

        RCE = GetRouteCacheEntry(Index, &Address);
        PrintRouteCacheEntry(RCE);
        free(RCE);
    }
    else {
        usage();
    }
}

IPV6_INFO_PREFIX_LIST *
GetPrefixListEntry(uint Index, IPv6Addr *Prefix, uint PrefixLength)
{
    IPV6_QUERY_PREFIX_LIST Query;
    IPV6_INFO_PREFIX_LIST *PLE;
    uint BytesReturned;

    Query.Index = Index;
    Query.Prefix = *Prefix;
    Query.PrefixLength = PrefixLength;

    PLE = (IPV6_INFO_PREFIX_LIST *) malloc(sizeof *PLE);
    if (PLE == NULL) {
        printf("malloc failed\n");
        exit(1);
    }

    if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_PREFIX_LIST,
                         &Query, sizeof Query,
                         PLE, sizeof *PLE, &BytesReturned,
                         NULL)) {
        printf("bad index or prefix\n");
        exit(1);
    }

    PLE->Query = Query;
    return PLE;
}

void
ForEachPrefix(void (*func)(IPV6_INFO_PREFIX_LIST *))
{
    IPV6_QUERY_PREFIX_LIST Query, NextQuery;
    IPV6_INFO_PREFIX_LIST PLE;
    uint BytesReturned;

    NextQuery.Index = 0;

    for (;;) {
        Query = NextQuery;

        if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_PREFIX_LIST,
                             &Query, sizeof Query,
                             &PLE, sizeof PLE, &BytesReturned,
                             NULL)) {
            printf("bad index %u\n", Query.Index);
            exit(1);
        }

        NextQuery = PLE.Query;

        if (Query.Index != 0) {

            PLE.Query = Query;
            (*func)(&PLE);
        }

        if (NextQuery.Index == 0)
            break;
    }
}

void
PrintPrefixListEntry(IPV6_INFO_PREFIX_LIST *PLE)
{
    printf("%u: %s/%u lifetime ",
           PLE->Query.Index,
           FormatIPv6Address(&PLE->Query.Prefix),
           PLE->Query.PrefixLength);

    if (PLE->ValidLifetime == (uint)-1)
        printf("infinite\n");
    else
        printf("%us\n", PLE->ValidLifetime);
}

void
QueryPrefixList(int argc, char *argv[])
{
    uint Index;
    IPv6Addr Prefix;
    uint PrefixLength;
    IPV6_INFO_PREFIX_LIST *PLE;

    if (argc == 0) {
        ForEachPrefix(PrintPrefixListEntry);
    }
    else if (argc == 2) {

        if ((Index = atoi(argv[0])) == 0)
            usage();

        if (! GetPrefix(argv[1], &Prefix, &PrefixLength))
            usage();

        PLE = GetPrefixListEntry(Index, &Prefix, PrefixLength);
        PrintPrefixListEntry(PLE);
        free(PLE);
    }
    else {
        usage();
    }
}

IPV6_INFO_ROUTER_LIST *
GetRouterListEntry(uint Index, IPv6Addr *Address)
{
    IPV6_QUERY_ROUTER_LIST Query;
    IPV6_INFO_ROUTER_LIST *RLE;
    uint BytesReturned;

    Query.Index = Index;
    Query.Address = *Address;

    RLE = (IPV6_INFO_ROUTER_LIST *) malloc(sizeof *RLE);
    if (RLE == NULL) {
        printf("malloc failed\n");
        exit(1);
    }

    if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_ROUTER_LIST,
                         &Query, sizeof Query,
                         RLE, sizeof *RLE, &BytesReturned,
                         NULL)) {
        printf("bad index or address\n");
        exit(1);
    }

    RLE->Query = Query;
    return RLE;
}

void
ForEachRouter(void (*func)(IPV6_INFO_ROUTER_LIST *))
{
    IPV6_QUERY_ROUTER_LIST Query, NextQuery;
    IPV6_INFO_ROUTER_LIST RLE;
    uint BytesReturned;

    NextQuery.Index = 0;

    for (;;) {
        Query = NextQuery;

        if (!DeviceIoControl(Handle, IOCTL_IPV6_QUERY_ROUTER_LIST,
                             &Query, sizeof Query,
                             &RLE, sizeof RLE, &BytesReturned,
                             NULL)) {
            printf("bad index %u\n", Query.Index);
            exit(1);
        }

        NextQuery = RLE.Query;

        if (Query.Index != 0) {

            RLE.Query = Query;
            (*func)(&RLE);
        }

        if (NextQuery.Index == 0)
            break;
    }
}

void
PrintRouterListEntry(IPV6_INFO_ROUTER_LIST *RLE)
{
    printf("%u: %s lifetime %us\n",
           RLE->Query.Index,
           FormatIPv6Address(&RLE->Query.Address),
           RLE->RouterLifetime);
}

void
QueryRouterList(int argc, char *argv[])
{
    uint Index;
    IPv6Addr Address;
    IPV6_INFO_ROUTER_LIST *RLE;

    if (argc == 0) {
        ForEachRouter(PrintRouterListEntry);
    }
    else if (argc == 2) {

        if ((Index = atoi(argv[0])) == 0)
            usage();

        if (! GetAddress(argv[1], &Address))
            usage();

        RLE = GetRouterListEntry(Index, &Address);
        PrintRouterListEntry(RLE);
        free(RLE);
    }
    else {
        usage();
    }
}
