// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Packet INternet Groper utility for IPv6.
//


#include <windows.h>
#include <devioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <winsock2.h>
#include <ws2ip6.h>
#include <ntddip6.h>


#define MAX_BUFFER_SIZE         (sizeof(ICMPV6_ECHO_REPLY) + 0xfff7)
#define DEFAULT_BUFFER_SIZE     (0x2000 - 8)
#define DEFAULT_SEND_SIZE       32
#define DEFAULT_COUNT           4
#define DEFAULT_TIMEOUT         1000L
#define MIN_INTERVAL            1000L

void
PrintUsage(void)
{
    printf("\nUsage: ping6 [-t] [-a] [-n count] [-l size]"
           " [-w timeout] [-s srcaddr] dest\n\n"
           "Options:\n"
           "-t             Ping the specifed host until interrupted.\n"
           "-a             Resolve addresses to hostnames.\n"
           "-n count       Number of echo requests to send.\n"
           "-l size        Send buffer size.\n"
           "-w timeout     Timeout in milliseconds to wait for each reply.\n"
           "-s srcaddr     Source address (required for link-local dest).\n");
}


int
get_pingee(char *ahstr, int dnsreq, struct in6_addr *address, char **hstr)
{
    struct hostent *hostp;

    if (inet6_addr(ahstr, address)) {
        // The user gave us a numeric IPv6 address.
        // Only do a reverse lookup if requested.

        if (dnsreq) {
            hostp = gethostbyaddr((char *)address, sizeof *address, AF_INET6);
        } else {
            hostp = NULL;
        }
    } else {
        // Presumably the user gave us a DNS name.

        hostp = getnodebyname(ahstr, AF_INET6, AI_DEFAULT);
        if ((hostp != NULL) && (hostp->h_length == sizeof *address)) {

            memcpy(address, hostp->h_addr, sizeof *address);
        } else {

            return FALSE;
        }
    }

    *hstr = (hostp != NULL) ? hostp->h_name : NULL;
    return TRUE;
}

int
get_source(char *astr, struct in6_addr *address)
{
    return inet6_addr(astr, address);
}

ulong
param(char **argv, int argc, int current, ulong min, ulong max)
{
    ulong   temp;
    char    *dummy;

    if (current == (argc - 1) ) {
        printf("Value must be supplied for option %s.\n", argv[current]);
        exit(1);
    }

    temp = strtoul(argv[current+1], &dummy, 0);
    if (temp < min || temp > max) {
        printf("Bad value for option %s.\n", argv[current]);
        exit(1);
    }

    return temp;
}

int _CRTAPI1
main(int argc, char **argv)
{
    char    *arg;
    uint    i;
    uint    j;
    int     found_addr = FALSE;
    int     dnsreq = FALSE;
    char    *hostname = NULL;
    struct in6_addr dstaddr, srcaddr;
    uint    Count = DEFAULT_COUNT;
    ulong   Timeout = DEFAULT_TIMEOUT;
    DWORD   errorCode;
    HANDLE  Handle;
    int     err;
    BOOL    result;
    PICMPV6_ECHO_REQUEST request;
    PICMPV6_ECHO_REPLY  reply;
    char    *SendBuffer, *RcvBuffer;
    uint    RcvSize;
    uint    SendSize = DEFAULT_SEND_SIZE;
    uint    ReplySize;
    DWORD   bytesReturned;
    WSADATA WsaData;

    memset(&srcaddr, 0, sizeof srcaddr);

    err = WSAStartup(MAKEWORD(2, 0), &WsaData);
    if (err) {
        printf("Unable to initialize Windows Sockets interface, error code %d\n", GetLastError());
        exit(1);
    }

    if (argc < 2) {
        PrintUsage();
        exit(1);
    } else {
        i = 1;
        while (i < (uint) argc) {
            arg = argv[i];
            if (arg[0] == '-') {        // Have an option
                switch (arg[1]) {
                case '?':
                    PrintUsage();
                    exit(0);

                case 'l':
                    // Avoid jumbo-grams, we don't support them yet.
                    // Need to allow 8 bytes for the Echo Request header.
                    SendSize = (uint)param(argv, argc, i++, 0, 0xffff - 8);
                    break;

                case 't':
                    Count = (uint)-1;
                    break;

                case 'n':
                    Count = (uint)param(argv, argc, i++, 1, 0xffffffff);
                    break;

                case 'w':
                    Timeout = param(argv, argc, i++, 0, 0xffffffff);
                    break;

                case 'a':
                    dnsreq = TRUE;
                    break;

                case 's':
                    if (!get_source(argv[++i], &srcaddr)) {
                        printf("Bad IPv6 address %s.\n", argv[i]);
                        exit(1);
                    }
                    break;

                default:
                    printf("Bad option %s.\n\n", arg);
                    PrintUsage();
                    exit(1);
                    break;
                }
                i++;
            } else {  // Not an option, must be an IPv6 address.
                if (found_addr) {
                    printf("Bad parameter %s.\n", arg);
                    exit(1);
                }
                if (get_pingee(arg, dnsreq, &dstaddr, &hostname)) {
                    found_addr = TRUE;
                    i++;
                } else {
                    printf("Bad IPv6 address %s.\n", arg);
                    exit(1);
                }
            }
        }
    }

    if (!found_addr) {
        printf("IPv6 address must be specified.\n");
        exit(1);
    }

    if (IN6_IS_ADDR_LINKLOCAL(&dstaddr) &&
        IN6_IS_ADDR_UNSPECIFIED(&srcaddr)) {
        printf("Link-local destination requires a source address\n"
               "to disambiguate the outgoing interface. Use -s.\n");
        exit(1);
    }

    Handle = CreateFileW(WIN_IPV6_DEVICE_NAME,
                         0,          // desired access
                         FILE_SHARE_READ | FILE_SHARE_WRITE,
                         NULL,       // security attributes
                         OPEN_EXISTING,
                         0,          // flags & attributes
                         NULL);      // template file
    if (Handle == INVALID_HANDLE_VALUE) {
        printf("Unable to contact IPv6 driver, error code %d.\n",
               GetLastError() );
        exit(1);
    }

    request = LocalAlloc(LMEM_FIXED, sizeof *request + SendSize);
    if (request == NULL) {
        printf("Unable to allocate required memory.\n");
        exit(1);
    }

    SendBuffer = (char *)(request + 1);

    //
    // Calculate receive buffer size and try to allocate it.
    //
    if (SendSize <= DEFAULT_SEND_SIZE) {
        RcvSize = DEFAULT_BUFFER_SIZE;
    }
    else {
        RcvSize = MAX_BUFFER_SIZE;
    }

    reply = LocalAlloc(LMEM_FIXED, sizeof *reply + RcvSize);
    if (reply == NULL) {
        printf("Unable to allocate required memory.\n");
        exit(1);
    }

    RcvBuffer = (char *)(reply + 1);

    //
    // Initialize the request buffer.
    //
    request->DstAddress = dstaddr;
    request->SrcAddress = srcaddr;
    request->Timeout = Timeout;
    request->TTL = 0; // default TTL

    //
    // Initialize the request data pattern.
    //
    for (i = 0; i < SendSize; i++)
        SendBuffer[i] = 'a' + (i % 23);

    if (hostname) {
        printf("\nPinging %s [%s] with %u bytes of data:\n\n",
               hostname, inet6_ntoa(&dstaddr), SendSize);
    } else {
        printf("\nPinging %s with %u bytes of data:\n\n",
               inet6_ntoa(&dstaddr), SendSize);
    }

    for (i = 0; i < Count; i++) {

        if (! DeviceIoControl(Handle,
                              IOCTL_ICMPV6_ECHO_REQUEST,
                              request,
                              sizeof *request + SendSize,
                              reply, sizeof *reply + RcvSize,
                              &bytesReturned,
                              NULL) ||
            (bytesReturned < sizeof *reply)) {

            errorCode = GetLastError();

            if (errorCode < IP_STATUS_BASE) {
                printf("PING: transmit failed, error code %u\n", errorCode);
            }
            else {
                printf("PING: %s\n", GetErrorString(errorCode));
            }
        }
        else {

            ReplySize = bytesReturned - sizeof *reply;

            if (! IN6_ADDR_EQUAL(&reply->Address, &in6addr_any))
                printf("Reply from %s: ", inet6_ntoa(&reply->Address));

            if (reply->Status == IP_SUCCESS) {

                printf("bytes=%u ", ReplySize);

                if (ReplySize != SendSize) {
                    printf("(sent %u) ", SendSize);
                }
                else {
                    char *sendptr, *recvptr;

                    sendptr = SendBuffer;
                    recvptr = RcvBuffer;

                    for (j = 0; j < SendSize; j++)
                        if (*sendptr++ != *recvptr++) {
                            printf("- MISCOMPARE at offset %u - ", j);
                            break;
                        }
                }

                if (reply->RoundTripTime) {
                    printf("time=%ums ", reply->RoundTripTime);
                }
                else {
                    printf("time<10ms ");
                }

                printf("\n");
            }
            else {
                printf("%s\n", GetErrorString(reply->Status));
            }

            if (i < (Count - 1)) {

                if (reply->RoundTripTime < MIN_INTERVAL) {
                    Sleep(MIN_INTERVAL - reply->RoundTripTime);
                }
            }
        }
    }

    return 0;
}
