// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Internet Control Message Protocol for Internet Protocol Version 6.
// See RFC 1885 for details.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "route.h"
#include "icmp.h"
#include <ntddip6.h>
#include "neighbor.h"

uint InfoMsgIntervalTicks;   // Min ticks between sends of ICMP info mesgs.
uint ErrorMsgIntervalTicks;  // Min ticks between sends of ICMP error mesgs.

// MLD external functions  BUGBUG: should be in an include file.
extern int
MLDQuery(NetTableEntry *RecvNTE, IPv6Packet *Packet,
         uchar Type, uchar Code);  

extern int
MLDReport(NetTableEntry *RecvNTE, IPv6Packet *Packet,
          uchar Type, uchar Code);  

extern MLDInit(void);


//
// Ping support.  We have a list of EchoControl blocks, one per outstanding
// echo request message.  Incoming echo replies are matched to requests via
// a unique sequence number.
//
KSPIN_LOCK ICMPv6EchoLock;
EchoControl *ICMPv6OutstandingEchos;
long ICMPv6EchoSeq;  // Protected with interlocked operations.



//* ICMPv6Init - Initialize ICMPv6.
//
//  Set the starting values of various things.
//
void
ICMPv6Init(void)
{
    //
    // Initialize in-kernel ping support.
    //
    ICMPv6OutstandingEchos = NULL;
    ICMPv6EchoSeq = 0;
    KeInitializeSpinLock(&ICMPv6EchoLock);

    InfoMsgIntervalTicks = ConvertMillisToTicks(ICMP_INFO_SEND_INTERVAL);
    ErrorMsgIntervalTicks = ConvertMillisToTicks(ICMP_ERROR_SEND_INTERVAL);

    //
    // Initialize Multicast Listener Discovery protocol.
    //
    MLDInit();
}

//* ICMPv6Send - Generic send routine for all ICMPv6 packets.
//
//  Common ICMPv6 message transmission functionality is performed here.
//  The message is expected to be completely formed (with the exception
//  of the checksum) when this routine is called.
//
//  REVIEW: Should we return an error code?  Let caller drop packets on error?
//
void
ICMPv6Send(
    NetTableEntry *NTE,   // NTE to send on (optional - NULL if don't care).
    PNDIS_PACKET Packet,  // Packet to send.
    uint IPv6Offset,      // Offset to IPv6 header in packet.
    uint ICMPv6Offset,    // Offset to ICMPv6 header in packet.
    IPv6Header *IP,       // Pointer to IPv6 header.
    uint PayloadLength,   // Length of IPv6 payload in bytes.
    ICMPv6Header *ICMP)   // Pointer to ICMPv6 header.
{
    RouteCacheEntry *RCE;
    IP_STATUS Status;
    uint ChecksumDataLength;
    LARGE_INTEGER Now;  // Time since boot in system interval timer ticks.
    LARGE_INTEGER LastSend;  
    ulong Interval;

    //
    // Get route.
    //
    Status = RouteToDestination(&IP->Dest, NTE, &RCE);
    if (Status != IP_SUCCESS) {
        //
        // No route - drop the packet.
        //
        DbgPrint("ICMPv6Send - no route: %x\n", Status);
        ASSERT(PC(Packet)->CompletionHandler == IPv6FreePacket);
        IPv6FreePacket(Packet);
        return;
    }

    if (NTE == NULL) {
        //
        // We were called without a specific NTE to send on.
        // This means we should take the source address from the RCE.
        //
        NTE = RCE->NTE;
        IP->Source = NTE->Address;
    }
    if (IP->HopLimit == 0)
        IP->HopLimit = NTE->IF->CurHopLimit;

    //
    // Calculate the ICMPv6 checksum.  It covers the entire ICMPv6 message
    // starting with the ICMPv6 header, plus the IPv6 pseudo-header.
    //
    // Recalculate the payload length to exclude any option headers.
    //
    ChecksumDataLength = PayloadLength - 
        (ICMPv6Offset - IPv6Offset) + sizeof(IPv6Header);

    ICMP->Checksum = 0;
    ICMP->Checksum = ChecksumPacket(Packet, ICMPv6Offset, NULL, 
                                    ChecksumDataLength,
                                    IP, IP_PROTOCOL_ICMPv6);

    // 
    // Limit the send rate of both informational and error ICMP messages.
    // Check the last time that the corresponding message was sent, if this
    // send is happening too soon, drop the packet.
    //
#if 0
    KeQueryTickCount(&Now);
    if (ICMPv6_INFORMATION_TYPE(ICMP->Type)) {
        if ((LONGLONG)(RCE->LastICMPInfoSend.QuadPart+InfoMsgIntervalTicks)-
            Now.QuadPart > 0) {
            goto too_fast;
        }
        KeQueryTickCount(&RCE->LastICMPInfoSend);  // reset timestamp.
    } else {
        if ((LONGLONG)(RCE->LastICMPErrorSend.QuadPart +
                       ErrorMsgIntervalTicks)- Now.QuadPart > 0) {
            goto too_fast;
        }
        KeQueryTickCount(&RCE->LastICMPErrorSend); // reset timestamp.
    }
#endif
    //
    // Hand the packet down to IP for transmission.
    //
    IPv6Send1(Packet, IPv6Offset, IP, PayloadLength, RCE);

    //
    // Release the route.
    //
    ReleaseRCE(RCE);
    return;

#if 0
  too_fast:
    DbgPrint("ICMPv6Send - rate limit violated.\n");
    ASSERT(PC(Packet)->CompletionHandler == IPv6FreePacket);
    IPv6FreePacket(Packet);
    return;
#endif
}


//* ICMPv6Reply - Generate a ICMPv6 message in response to an incoming packet.
// 
//  Send an ICMPv6 message of the given Type and Code to the source of the
//  offending/invoking packet.  The reply includes as much of the incoming
//  packet as will fit inside the minimal IPv6 MTU.
//
int
ICMPv6Reply(
    NetTableEntry *RecvNTE,     // NTE packet was received on.
    IPv6Packet *Packet,         // Packet handed to us by link layer.
    ICMPv6MsgType *ICMPType)   // Specifies msg type, code, and data pointer. 
{
    Interface *IF = RecvNTE->IF;
    NDIS_STATUS Status;
    PNDIS_PACKET ReplyPacket;
    PNDIS_BUFFER Buffer;
    uint Offset;
    uchar *Mem;
    int MemLen;
    uint ICMPLength;
    uint PayloadLength;
    IPv6Header *ReplyIP;
    ICMPv6Header *ICMP;

    //
    // Basically what we do here is slap an ICMPv6 header on the front
    // of the invoking packet and send it back where it came from.
    //
    //
    // Allocate a header for reply packet.
    //
    NdisAllocatePacket(&Status, &ReplyPacket, IPv6PacketPool);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("ICMPv6: Couldn't allocate packet header!?!\n");
        return 0;
    }
    PC(ReplyPacket)->CompletionHandler = IPv6FreePacket;

    //
    // Calculate the length of the ICMP header
    // and how much data we will include following the ICMP header.
    //
    if (ICMPv6_ERROR_TYPE(ICMPType->Type)) {
        //
        // Include space for an error value after the header proper.
        //
        ICMPLength = sizeof(ICMPv6Header) + sizeof(uint);
        //
        // We want to include data from the IP header on.
        //
        PayloadLength = Packet->TotalSize +
            (Packet->Position - Packet->IPPosition);
        //
        // But limit the error packet size.
        //
        if (PayloadLength > ICMPv6_ERROR_MAX_DATA_LEN)
            PayloadLength = ICMPv6_ERROR_MAX_DATA_LEN;
    } else {
        //
        // Use a vanilla ICMP header.
        // Include only remaining data from the packet.
        //
        ICMPLength = sizeof(ICMPv6Header);
        PayloadLength = Packet->TotalSize;
    }

    //
    // Allocate memory for buffer.
    // BUGBUG: We're assuming here that there are no intermediate
    // BUGBUG: headers we'll need to put in our Echo Reply message.
    //
    Offset = MaxLinkLevelHeader;
    MemLen = Offset + sizeof(IPv6Header) + ICMPLength + PayloadLength;
    Mem = ExAllocatePool(NonPagedPool, MemLen);
    if (Mem == NULL) {
        DbgPrint("ICMPv6: Couldn't allocate memory!?!\n");
        return 0;
    }

    //
    // Allocate a buffer header and put it on the packet chain.
    //
    NdisAllocateBuffer(&Status, &Buffer, IPv6BufferPool, Mem, MemLen);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("ICMPv6: Couldn't allocate buffer header!?!\n");
        return 0;
    }
    NdisChainBufferAtFront(ReplyPacket, Buffer);

    //
    // Prepare IP header of reply packet.
    //
    // REVIEW: Can/should some of this be moved to ICMPv6Send or IPv6Send?
    // REVIEW: We use incoming Traffic Class for outbound Traffic Class.
    //
    ReplyIP = (IPv6Header *)(Mem + Offset);
    ReplyIP->VersClassFlow = Packet->IP->VersClassFlow &
        (IP_VER_MASK | IP_TRAFFIC_CLASS_MASK);
    ReplyIP->NextHeader = IP_PROTOCOL_ICMPv6;
    ReplyIP->HopLimit = 0; // ICMPv6Send will fill in the default.

    //
    // Take our reply's destination address from the source address
    // of the incoming packet.
    //
    // REVIEW: What do we need to do to support source routed packets?
    //
    // REVIEW: Catch non-unique source addresses here?  We MUST NOT
    // REVIEW: reply to messages from the unspecified address, a
    // REVIEW: multicast address or a known anycast address.
    //
    ReplyIP->Dest = Packet->IP->Source;

    //
    // Take our reply's source address from the receiving NTE.
    //
    ReplyIP->Source = RecvNTE->Address;

    //
    // Prepare ICMP header.
    //
    // REVIEW: Do this in ICMPv6Send?
    //
    ICMP = (ICMPv6Header *)(ReplyIP + 1);
    ICMP->Type = ICMPType->Type;
    ICMP->Code = ICMPType->Code;

    if (ICMPv6_INFORMATION_TYPE(ICMPType->Type)) {
        //
        // This is an ICMP Information message - an echo reply.
        //
        // Copy incoming packet data to outgoing.
        //
        // BUGBUG: This code is copying from ICMP Header onward (and assumes
        // BUGBUG: that the data is in a single flat buffer.)
        //
        RtlCopyMemory(ICMP + 1, Packet->Data, PayloadLength);
        
    } else {
        //
        // This is an ICMP Error Message.  
        //
        // For some error massages the error value specifies the offset
        // into data packet where the problem occured. This is placed
        // just after the ICMP header proper.
        //
        *(uint *)(ICMP + 1) = net_long(ICMPType->Value);

        //
        // Copy invoking packet data to outgoing.
        //
        // BUGBUG: This code is copying from the invoking packet's IP 
        // BUGBUG: header onward (and assumes that the data is in a 
        // BUGBUG: single flat buffer.)
        //
        RtlCopyMemory((uint *)(ICMP + 1) + 1, Packet->IP, PayloadLength);
    }

    ICMPv6Send(RecvNTE, ReplyPacket, Offset,
               Offset + sizeof(IPv6Header), ReplyIP,
               ICMPLength + PayloadLength, ICMP);
    return 0;
}


//* ICMPv6EchoReply
//
//  Called either when an echo reply arrives, or
//  a hop-count-exceeded error responding to an echo request arrives.
//
//  Looks up the echo request structure and completes
//  the echo request operation.
//
void
ICMPv6EchoReply(
    ulong Seq,                  // Echo sequence number.
    IP_STATUS Status,           // Status of the response.
    IPv6Addr *Address,          // Source of the reply.
    void *Current,              // Pointer to the buffered data area.
    uint PayloadLength)         // Size of remaining payload data.
{
    EchoControl *This, **PrevPtr;

    //
    // Find the EchoControl block on our list of outstanding echos that
    // has a matching sequence number and call it's completion function.
    //
    KeAcquireSpinLockAtDpcLevel(&ICMPv6EchoLock);
    for (This = ICMPv6OutstandingEchos, PrevPtr = &ICMPv6OutstandingEchos;
         This != (EchoControl *)NULL; This = This->Next) {
        if (This->Seq == Seq) {
            //
            // Found matching control block.  Extract it from list.
            //
            *PrevPtr = This->Next;
            break;
        }
        PrevPtr = &This->Next;
    }
    KeReleaseSpinLockFromDpcLevel(&ICMPv6EchoLock);

    //
    // Check to see if we ran off the end of the outstanding echos list.
    //
    if (This == NULL) {
        //
        // We received a response with a sequence number that doesn't match
        // one of the echo requests we've sent out.  Drop it.
        //
        DbgPrint("ICMPv6: Echo response with bogus sequence number %x\n",
                 Seq);
        return;
    }

    //
    // Call OS-specific completion routine.
    //
    // BUGBUG: This currently assumes all the echo data fits into the
    // BUGBUG: current data buffer.
    //
    (*This->CompleteRoutine)(This, Status, Address, Current, PayloadLength);
}

//* ICMPv6EchoReplyReceive - Receive a reply to an earlier echo of our's.
//
//  Called by ICMPv6Receive when an echo reply message arrives.
//
//  REVIEW: Should we also verify the receiving NTE is the same as the one
//  REVIEW: we sent on?
//
int
ICMPv6EchoReplyReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by link layer.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code)                 // Value of Code field in ICMPv6 header.
{
    ulong Seq;

    if (Packet->TotalSize < sizeof Seq) {
        DbgPrint("ICMPv6: Received small Echo Reply %u\n", 
                 Packet->TotalSize);
        return 0;
    }
    
    //
    // We're received a reply message to one of our echo requests.
    // Extract its sequence number so that we can identify it.
    //
    Seq = net_long(*(ulong *)Packet->Data);
    (uchar *)Packet->Data += sizeof Seq;
    Packet->TotalSize -= sizeof Seq;
    
    ICMPv6EchoReply(Seq, IP_SUCCESS, &Packet->IP->Source, 
                    Packet->Data, Packet->TotalSize);
    return 0;
}


//* ICMPv6PacketTooBig - Receive a Packet Too Big ICMP message.
//
//  Parses the packet and then calls the route cache
//  to perform the actual update of the path MTU.
//
void
ICMPv6PacketTooBig(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet)         // Packet handed to us by link layer.
{
    IPv6Header UNALIGNED *InvokingIP;
    uint MTU;

    //
    // The packet should contain a full Packet Too Big message,
    // plus enough of the invoking packet to include the header.
    //
    if (Packet->TotalSize < sizeof MTU + sizeof *InvokingIP) {
        DbgPrint("ICMPv6: Received small Packet Too Big %u\n", 
                 Packet->TotalSize);
        return;
    }

    //
    // Pull the new MTU out of the packet.
    //
    MTU = net_long(*(ulong *)Packet->Data);
    if (MTU < IPv6_MINIMUM_MTU) {
        DbgPrint("ICMPv6: Received Packet Too Big with bogus MTU %u\n", MTU);
        return;
    }

    //
    // The "invoking packet" follows the MTU.
    // We've already checked that the entire header is present.
    //
    InvokingIP = (IPv6Header UNALIGNED *)((ulong *)Packet->Data + 1);

    //
    // Update the path MTU in the route cache.
    //
    UpdatePathMTU(&InvokingIP->Dest, RecvNTE, MTU);
}


//* ICMPv6ErrorReceive
//
//  Called by ICMPv6Receive when an error message arrives.
//
//  BUGBUG: Currently only checks for outstanding echo requests.
//  We should pass all errors to upper layers.
//
int
ICMPv6ErrorReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    IP_STATUS Status)           // ICMP Error Code.
{
    ulong Parameter;
    StatusArg StatArg;
    IPv6Header *InvokingIP;
    ProtoControlRecvProc *CtlProc = NULL;
    int PacketRef;

    //
    // All ICMPv6 error messages consist of the base ICMPv6 header,
    // followed by a 32 bit type-specific field, followed by as much
    // of the invoking packet as fit without causing this ICMPv6 packet
    // to exceed 576 octets.
    //
    // We already consumed the base ICMPv6 header back in ICMPv6Receive.
    // Pull out the 32 bit type-specific field in case the upper layer's
    // ControlReceive routine cares about it.
    //
    if (Packet->ContigSize < sizeof(ulong)) {
        if (PacketPullup(Packet, sizeof(ulong)) == 0) {
            // Pullup failed.
            DbgPrint("ICMPv6: Packet too small to contain error field\n");
            return 0;  // Drop packet.
        }
    }
    Parameter = *(ulong *)Packet->Data;
    (uchar *)Packet->Data += sizeof(ulong);
    Packet->ContigSize -= sizeof(ulong);
    Packet->TotalSize -= sizeof(ulong);
    Packet->Position += sizeof(ulong);

    //
    // Next up should be the IPv6 header of the invoking packet.
    //
    if (Packet->ContigSize < sizeof(IPv6Header)) {
        if (PacketPullup(Packet, sizeof(IPv6Header)) == 0) {
            // Pullup failed.
            DbgPrint("ICMPv6: Packet too small to contain IPv6 header "
                     "from invoking packet\n");
            return 0;  // Drop packet.
        }
    }
    InvokingIP = (IPv6Header UNALIGNED *)Packet->Data;
    (uchar *)Packet->Data += sizeof(IPv6Header);
    Packet->ContigSize -= sizeof(IPv6Header);
    Packet->TotalSize -= sizeof(IPv6Header);
    Packet->Position += sizeof(IPv6Header);

    //
    // Deliver ICMP Error to higher layers.
    //
    // BUGBUG: This should probably loop through each extension header
    // BUGBUG: rather than have each call the next recursively.
    //
    CtlProc = ProtocolSwitchTable[InvokingIP->NextHeader].ControlReceive;

    if (CtlProc) {
        //
        // BUGBUG: Should pass the Parameter we pulled out above.
        //
        StatArg.Status = Status;
        StatArg.Arg = Parameter;
        StatArg.Other = InvokingIP->NextHeader;
        PacketRef = (*CtlProc)(RecvNTE, Packet, &StatArg);
        return PacketRef;
    }

    return 0; // discard error packet.
}

//* ICMPv6ControlReceive - handler for ICMPv6 control messages.
//
//  This routine is called if we receive an ICMPv6 error message that
//  was generated by some remote site as a result of receiving an ICMPv6
//  packet from us.
//  
ICMPv6ControlReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    StatusArg *StatArg)         // ICMP Error Code and pointer value.
{
    ICMPv6Header *InvokingICMP;
    ulong Seq;

    //
    // The next thing in the packet should be the ICMP header of the
    // original packet which invoked this error.
    //
    if (Packet->ContigSize < sizeof(ICMPv6Header)) {
        if (PacketPullup(Packet, sizeof(ICMPv6Header)) == 0) {
            // Pullup failed.
            DbgPrint("ICMPv6: Packet too small to contain ICMPv6 header "
                     "from invoking packet\n");
            return 0;  // Drop packet.
        }
    }
    InvokingICMP = (ICMPv6Header UNALIGNED *)Packet->Data;

    //
    // BUGBUG: All we currently handle is errors caused by echo requests.
    //
    if ((InvokingICMP->Type != ICMPv6_ECHO_REQUEST) ||
        (InvokingICMP->Code != 0))
        return 0;

    Seq = net_long(*(ulong UNALIGNED *)(InvokingICMP + 1));

    //
    // Complete the corresponding echo request with an error.
    //
    ICMPv6EchoReply(Seq, StatArg->Status, &Packet->IP->Source, NULL, 0);
    return 0;
}


//* ICMPv6Receive - Receive an incoming ICMPv6 packet.
// 
//  This is the routine called by IPv6 when it receives a complete IPv6
//  packet with a Next Header value of 58.
//
int
ICMPv6Receive(
    NetTableEntry *RecvNTE,     // NTE that we received the packet on.
    IPv6Packet *Packet)         // Packet handed to us by link layer.
{
    ICMPv6Header UNALIGNED *ICMP;
    ushort Checksum;

    //
    // Verify checksum.
    //
    Checksum = ChecksumPacket(Packet->NdisPacket, Packet->Position,
                              Packet->Data, Packet->TotalSize,
                              Packet->IP, IP_PROTOCOL_ICMPv6);
    if (Checksum != 0xffff) {
        DbgPrint("ICMPv6: Checksum failed %0x\n", Checksum);
        return 0;  // Drop packet.
    }

    //
    // Verify that we have enough contiguous data to overlay a ICMPv6Header
    // structure on the incoming packet.  Then do so.
    //
    if (Packet->ContigSize < sizeof(ICMPv6Header)) {
        if (PacketPullup(Packet, sizeof(ICMPv6Header)) == 0) {
            // Pullup failed.
            DbgPrint("ICMPv6: Packet too small to contain ICMP header\n");
            return 0;  // Drop packet.
        }
    }
    ICMP = (ICMPv6Header UNALIGNED *)Packet->Data;
    //
    // Skip over base ICMP header.
    //
    (uchar *)Packet->Data += sizeof *ICMP;
    Packet->ContigSize -= sizeof *ICMP;
    Packet->TotalSize -= sizeof *ICMP;
    Packet->Position += sizeof *ICMP;

    //
    // Check that the receiving NTE is valid
    // (invalid NTEs are typically due to DAD),
    // unless this is a NS or NA (DAD uses them)
    // or multicast RA (want to receive during DAD).
    //
    if (!IsValidNTE(RecvNTE)) {
        if ((ICMP->Type == ICMPv6_NEIGHBOR_SOLICIT) ||
            (ICMP->Type == ICMPv6_NEIGHBOR_ADVERT) ||
            ((ICMP->Type == ICMPv6_ROUTER_ADVERT) &&
             IP6_ADDR_EQUAL(&Packet->IP->Dest, &AllNodesOnLinkAddr))) {
            //
            // Receive this packet anyway.
            //
        } else {
            DbgPrint("ICMPv6Receive: invalid NTE\n");
            return 0; // Drop packet.
        }
    }

    //
    // Ignore Neighbor Discovery packets
    // if the interface is so configured.
    // (Pseudo-interfaces don't do Neighbor Discovery.)
    //
    if (RecvNTE->IF->DisableND) {
        if ((ICMP->Type == ICMPv6_ROUTER_SOLICIT) ||
            (ICMP->Type == ICMPv6_ROUTER_ADVERT) ||
            (ICMP->Type == ICMPv6_NEIGHBOR_SOLICIT) ||
            (ICMP->Type == ICMPv6_NEIGHBOR_ADVERT) ||
            (ICMP->Type == ICMPv6_REDIRECT)) {

            DbgPrint("ICMPv6Receive: ND on pseudo-interface\n");
            return 0; // Drop packet.
        }
    }

    //
    // Handle specific message types.
    // Just use a switch statement for now.  If this is later deemed to be
    // too inefficient, we can change it to use a type switch table instead.
    //
    if (ICMPv6_ERROR_TYPE(ICMP->Type)) {
        IP_STATUS Status;

        //
        // Pull-up code should guarantee that we have contiguous data.
        //
        ASSERT(Packet->ContigSize == Packet->TotalSize);

        //
        // First we perform any specific processing of the error,
        // and convert the error type/code to a status value.
        //
        switch (ICMP->Type) {
        case ICMPv6_DESTINATION_UNREACHABLE:
            switch (ICMP->Code) {
            case ICMPv6_NO_ROUTE_TO_DESTINATION:
                Status = IP_DEST_NO_ROUTE;
                break;
            case ICMPv6_COMMUNICATION_PROHIBITED:
                Status = IP_DEST_PROHIBITED;
                break;
            case ICMPv6_NOT_NEIGHBOR:
                Status = IP_DEST_NOT_NEIGHBOR;
                break;
            case ICMPv6_ADDRESS_UNREACHABLE:
                Status = IP_DEST_ADDR_UNREACHABLE;
                break;
            case ICMPv6_PORT_UNREACHABLE:
                Status = IP_DEST_PORT_UNREACHABLE;
                break;
            default:
                Status = IP_DEST_UNREACHABLE;
                break;
            }
            break;

        case ICMPv6_PACKET_TOO_BIG:
            if (ICMP->Code == 0)
                ICMPv6PacketTooBig(RecvNTE, Packet);

            Status = IP_PACKET_TOO_BIG;
            break;

        case ICMPv6_TIME_EXCEEDED:
            switch (ICMP->Code) {
            case ICMPv6_HOP_LIMIT_EXCEEDED:
                Status = IP_HOP_LIMIT_EXCEEDED;
                break;
            case ICMPv6_REASSEMBLY_TIME_EXCEEDED:
                Status = IP_REASSEMBLY_TIME_EXCEEDED;
                break;
            default:
                Status = IP_TIME_EXCEEDED;
                break;
            }
            break;

        case ICMPv6_PARAMETER_PROBLEM:
            switch (ICMP->Code) {
            case ICMPv6_ERRONEOUS_HEADER_FIELD:
                Status = IP_BAD_HEADER;
                break;
            case ICMPv6_UNRECOGNIZED_NEXT_HEADER:
                Status = IP_UNRECOGNIZED_NEXT_HEADER;
                break;
            case ICMPv6_UNRECOGNIZED_OPTION:
                Status = IP_BAD_OPTION;
                break;
            default:
                Status = IP_PARAMETER_PROBLEM;
                break;
            }
            break;
            
        default:
            Status = IP_ICMP_ERROR;
            break;
        }

        //
        // Now we perform generic error processing.
        // We MUST notify upper layers, even if we don't recognize
        // the specific error message.
        //
        return ICMPv6ErrorReceive(RecvNTE, Packet, Status);
        
    } else {
        ICMPv6MsgType ICMPType;
        switch(ICMP->Type) {
        case ICMPv6_ECHO_REQUEST:
            ICMPType.Type = (uchar) ICMPv6_ECHO_REPLY;
            ICMPType.Code = 0;
            ICMPType.Value = 0;
            return ICMPv6Reply(RecvNTE, Packet, &ICMPType);

        case ICMPv6_ECHO_REPLY:
            return ICMPv6EchoReplyReceive(RecvNTE, Packet, 
                                          ICMP->Type, ICMP->Code);

        case ICMPv6_MULTICAST_LISTENER_QUERY:
            return MLDQuery(RecvNTE, Packet,
                            ICMP->Type, ICMP->Code);

        case ICMPv6_MULTICAST_LISTENER_REPORT:
            return MLDReport(RecvNTE, Packet,
                             ICMP->Type, ICMP->Code);

        case ICMPv6_MULTICAST_LISTENER_DONE:
            DbgPrint("ICMPv6: Got a Group Membership Reduction message\n");
            break;


        // Following are all Neighbor Discovery messages.
        case ICMPv6_ROUTER_SOLICIT:
            return RouterSolicitReceive(RecvNTE, Packet,
                                        ICMP->Type, ICMP->Code);

        case ICMPv6_ROUTER_ADVERT:
            return RouterAdvertReceive(RecvNTE, Packet,
                                       ICMP->Type, ICMP->Code);

        case ICMPv6_NEIGHBOR_SOLICIT:
            return NeighborSolicitReceive(RecvNTE, Packet, 
                                          ICMP->Type, ICMP->Code);

        case ICMPv6_NEIGHBOR_ADVERT:
            return NeighborAdvertReceive(RecvNTE, Packet, 
                                         ICMP->Type, ICMP->Code);

        case ICMPv6_REDIRECT:
            return RedirectReceive(RecvNTE, Packet,
                                   ICMP->Type, ICMP->Code);

        default:
            //
            // Don't recognize the specific message type.
            // This is an informational message.
            // We MUST silently discard it.
            //
            KdPrint(("ICMPv6: Received unknown informational message"
                     "(%u/%u) from %ws\n", ICMP->Type, ICMP->Code,
                     FormatV6Address(&Packet->IP->Source)));
            break;
        }
    }
    return 0;
}


//* ICMPv6EchoRequest - Common dispatch routine for echo requests.
//
//  This is the routine called by the OS-specific code on behalf of a user
//  to issue an echo request.  Validate the request, place control block
//  on list of outstanding echo requests, and send echo request message.
//
IP_STATUS                       // Returns: IP_STATUS of attempt to ping.
ICMPv6EchoRequest(
    void *InputBuffer,          // Pointer to an ICMPV6_ECHO_REQUEST structure.
    uint InputBufferLength,     // Size in bytes of the InputBuffer.
    EchoControl *ControlBlock,  // Pointer to an EchoControl structure.
    EchoRtn Callback)           // Called when request responds or times out.
{
    NetTableEntry *NTE = NULL;
    PICMPV6_ECHO_REQUEST RequestBuffer;
    KIRQL OldIrql;
    IP_STATUS Status;
    ulong Seq;
    NDIS_STATUS NdisStatus;
    PNDIS_PACKET Packet;
    PNDIS_BUFFER Buffer;
    uint Offset;
    uchar *Mem;
    int MemLen;
    IPv6Header *IP;
    ICMPv6Header *ICMP;
    void *Data;
    uint DataSize;

    RequestBuffer = (PICMPV6_ECHO_REQUEST) InputBuffer;

    //
    // Validate the request.
    //
    if (InputBufferLength < sizeof *RequestBuffer) {
        Status = IP_BUF_TOO_SMALL;
        goto common_echo_exit;
    }

    Data = RequestBuffer + 1;
    DataSize = InputBufferLength - sizeof *RequestBuffer;

    //
    // Determine which NTE will send the request,
    // if the user has specified a source address.
    //
    if (! IsUnspecified(&RequestBuffer->SrcAddress)) {
        //
        // Convert the source address to an NTE.
        //
        NTE = FindNetworkWithAddress(&RequestBuffer->SrcAddress);
        if (NTE == NULL) {
            Status = IP_BAD_ROUTE;
            goto common_echo_exit;
        }
    }

    //
    // Allocate a header for Echo Request packet.
    //
    NdisAllocatePacket(&NdisStatus, &Packet, IPv6PacketPool);
    if (NdisStatus != NDIS_STATUS_SUCCESS) {
        Status = IP_NO_RESOURCES;
        goto common_echo_exit;
    }
    PC(Packet)->CompletionHandler = IPv6FreePacket;

    //
    // Allocate memory for buffer.
    //
    Offset = MaxLinkLevelHeader;
    MemLen = Offset + sizeof(IPv6Header) + sizeof(ICMPv6Header) +
        sizeof Seq + DataSize;
    Mem = ExAllocatePool(NonPagedPool, MemLen);
    if (Mem == NULL) {
        Status = IP_NO_RESOURCES;
        goto common_echo_exit;
    }

    //
    // Allocate a buffer header and put it on the packet chain.
    //
    NdisAllocateBuffer(&NdisStatus, &Buffer, IPv6BufferPool, Mem, MemLen);
    if (NdisStatus != NDIS_STATUS_SUCCESS) {
        Status = IP_NO_RESOURCES;
        goto common_echo_exit;
    }
    NdisChainBufferAtFront(Packet, Buffer);

    //
    // Prepare IP header of Echo Request packet.
    //
    IP = (IPv6Header *)(Mem + Offset);
    IP->VersClassFlow = IP_VERSION;
    IP->NextHeader = IP_PROTOCOL_ICMPv6;
    IP->Dest = RequestBuffer->DstAddress;
    if (NTE != NULL)
        IP->Source = NTE->Address;
    IP->HopLimit = RequestBuffer->TTL;

    //
    // Prepare ICMP header.
    //
    ICMP = (ICMPv6Header *)(IP + 1);
    ICMP->Type = ICMPv6_ECHO_REQUEST;
    ICMP->Code = 0;

    //
    // Insert Echo sequence number.  Technically, this is 16 bits of
    // "Identifier" and 16 bits of "Sequence Number", but we just treat
    // the whole thing as one 32 bit sequence number field.
    //
    Seq = InterlockedIncrement(&ICMPv6EchoSeq);
    Mem = (uchar *)(ICMP + 1);
    *(ulong *)Mem = net_long(Seq);
    Mem += sizeof(ulong);

    //
    // Copy the user data into the packet.
    //
    RtlCopyMemory(Mem, Data, DataSize);

    //
    // Prepare the control block and link it onto the list.
    // Once we've unlocked the list, the control block might
    // be completed at any time.  Hence it's very important
    // that we not access RequestBuffer after this point.
    //
    ControlBlock->TimeoutTimer = ConvertMillisToTicks(RequestBuffer->Timeout);
    ControlBlock->CompleteRoutine = Callback;
    ControlBlock->Seq = Seq;

    KeAcquireSpinLock(&ICMPv6EchoLock, &OldIrql);
    ControlBlock->Next = ICMPv6OutstandingEchos;
    ICMPv6OutstandingEchos = ControlBlock;
    KeReleaseSpinLock(&ICMPv6EchoLock, OldIrql);

    ICMPv6Send(NTE, Packet, Offset, Offset + sizeof(IPv6Header), IP,
               sizeof(ICMPv6Header) + sizeof Seq + DataSize, ICMP);

    Status = IP_PENDING;

common_echo_exit:

    if (NTE != NULL)
        ReleaseNTE(NTE);

    return Status;

} // ICMPv6EchoRequest


//* ICMPv6EchoComplete - Common completion routine for echo requests.
//
//  This is the routine is called by the OS-specific code to process an
//  ICMP echo response.
//
NTSTATUS
ICMPv6EchoComplete(
    EchoControl *ControlBlock,  // ControlBlock of completed request.
    IP_STATUS Status,           // Status of the reply.
    IPv6Addr *Address,          // Source of the reply.
    void *Data,                 // Reply data (may be NULL).
    uint DataSize,              // Amount of reply data.
    ULONG *BytesReturned)       // Total user bytes returned.
{
    PICMPV6_ECHO_REPLY ReplyBuffer;
    LARGE_INTEGER Now;

    //
    // Sanity check our reply buffer length.
    //
    if (ControlBlock->ReplyBufLen < sizeof *ReplyBuffer) {
        *BytesReturned = 0;
        return STATUS_BUFFER_TOO_SMALL;
    }

    ReplyBuffer = (PICMPV6_ECHO_REPLY) ControlBlock->ReplyBuf;

    //
    // Fill in fields to return.
    //
    ReplyBuffer->Address = *Address;
    ReplyBuffer->Status = Status;
    KeQueryTickCount(&Now);
    ReplyBuffer->RoundTripTime = (unsigned int)((KeQueryTimeIncrement() * 
        (Now.QuadPart - ControlBlock->WhenIssued.QuadPart)) / 10000);

    //
    // Verify we have enough space in the reply buffer for the reply data.
    //
    if (ControlBlock->ReplyBufLen < sizeof *ReplyBuffer + DataSize) {
        *BytesReturned = sizeof *ReplyBuffer;
        return STATUS_BUFFER_TOO_SMALL;
    }

    //
    // Copy the reply data to follow the reply buffer.
    //
    RtlCopyMemory(ReplyBuffer + 1, Data, DataSize);

    *BytesReturned = sizeof *ReplyBuffer + DataSize;
    return STATUS_SUCCESS;

}  // ICMPv6EchoComplete


//* ICMPv6EchoTimeout - expire aging unanswered echo requests.
//
//  IPv6Timeout calls this routine whenever it thinks we might have
//  echo requests outstanding.
//
void
ICMPv6EchoTimeout(void)
{
    KIRQL OldIrql;
    EchoControl *This, **PrevPtr, *TimedOut;

    TimedOut = NULL;

    //
    // Grab the outstanding echo list lock and run through the list looking
    // for requests that have timed out.
    //
    KeAcquireSpinLock(&ICMPv6EchoLock, &OldIrql);
    for (This = ICMPv6OutstandingEchos, PrevPtr = &ICMPv6OutstandingEchos;
         This != (EchoControl *)NULL; This = This->Next) {
        if (This->TimeoutTimer != 0) {
            //
            // Timer is running.  Decrement and check for expiration.
            //
            if (--This->TimeoutTimer == 0) {
                //
                // This echo request has been sent and timed out without
                // being answered.  Move it to our timed out list.
                //
                *PrevPtr = This->Next;
                This->Next = TimedOut;
                TimedOut = This;
            } else {
                PrevPtr = &This->Next;
            }
        }
    }
    KeReleaseSpinLock(&ICMPv6EchoLock, OldIrql);

    //
    // Run through the list of timed out echos, calling the completion
    // routine on each.  The completion routine is responsible for
    // freeing the EchoControl block structure.
    //
    while (TimedOut != NULL) {
        
        This = TimedOut;
        TimedOut = This->Next;

        (*This->CompleteRoutine)(This, IP_REQ_TIMED_OUT,
                                 &UnspecifiedAddr, NULL, 0);
    }
}
