// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Internet Control Message Protocol for IPv6 definitions.
// See RFC 1885 and RFC 1970 for details.
//


#ifndef ICMPv6_INCLUDED
#define ICMPv6_INCLUDED 1

//
// ICMPv6 Header.
// The actual message body follows this header and is type-specific.
//
typedef struct ICMPv6Header {
    uchar Type;       // Type of message (high bit zero for error messages).
    uchar Code;       // Type-specific differentiater.
    ushort Checksum;  // Calculated over ICMPv6 message and IPv6 psuedo-header.
} ICMPv6Header;


//
// ICMPv6 Type field definitions.
//
#define ICMPv6_DESTINATION_UNREACHABLE    1
#define ICMPv6_PACKET_TOO_BIG             2
#define ICMPv6_TIME_EXCEEDED              3
#define ICMPv6_PARAMETER_PROBLEM          4

#define ICMPv6_ECHO_REQUEST               128
#define ICMPv6_ECHO_REPLY                 129
#define ICMPv6_MULTICAST_LISTENER_QUERY   130
#define ICMPv6_MULTICAST_LISTENER_REPORT  131
#define ICMPv6_MULTICAST_LISTENER_DONE    132

#define ICMPv6_ROUTER_SOLICIT             133
#define ICMPv6_ROUTER_ADVERT              134
#define ICMPv6_NEIGHBOR_SOLICIT           135
#define ICMPv6_NEIGHBOR_ADVERT            136
#define ICMPv6_REDIRECT                   137

#define ICMPv6_INFORMATION_TYPE(type)   ((type) & 0x80)
#define ICMPv6_ERROR_TYPE(type)         (((type) & 0x80) == 0)

// Max amount of packet data in an ICMP error message is: 
//      576 - IP header - ICMP header - ICMP Pointer = 528.
#define ICMPv6_ERROR_MAX_DATA_LEN       528   


//
// ICMPv6 Code field definitions.
//

// For Destination Unreachable errors:
#define ICMPv6_NO_ROUTE_TO_DESTINATION          0
#define ICMPv6_COMMUNICATION_PROHIBITED         1
#define ICMPv6_NOT_NEIGHBOR                     2
#define ICMPv6_ADDRESS_UNREACHABLE              3
#define ICMPv6_PORT_UNREACHABLE                 4

// For Time Exceeded errors:
#define ICMPv6_HOP_LIMIT_EXCEEDED               0
#define ICMPv6_REASSEMBLY_TIME_EXCEEDED         1

// For Parameter Problem errors:
#define ICMPv6_ERRONEOUS_HEADER_FIELD           0
#define ICMPv6_UNRECOGNIZED_NEXT_HEADER         1
#define ICMPv6_UNRECOGNIZED_OPTION              2

//
//  Neighbor Discovery message option definitions.
//
#define ND_OPTION_SOURCE_LINK_LAYER_ADDRESS 1
#define ND_OPTION_TARGET_LINK_LAYER_ADDRESS 2
#define ND_OPTION_PREFIX_INFORMATION 3
#define ND_OPTION_REDIRECTED_HEADER 4
#define ND_OPTION_MTU 5

typedef struct NDOptionMTU {
    uchar Type;
    uchar Length;
    ushort Reserved;
    uint MTU;
} NDOptionMTU;

typedef struct NDOptionPrefixInformation {
    uchar Type;
    uchar Length;
    uchar PrefixLength;
    uchar Flags;
    uint ValidLifetime;
    uint PreferredLifetime;
    uint Reserved2;
    IPv6Addr Prefix;
} NDOptionPrefixInformation;

//
//  Neighbor Advertisement message flags.
//
#define ND_NA_FLAG_ROUTER    0x80000000
#define ND_NA_FLAG_SOLICITED 0x40000000
#define ND_NA_FLAG_OVERRIDE  0x20000000

//
// Router Advertisement message flags.
//
#define ND_RA_FLAG_MANAGED   0x80
#define ND_RA_FLAG_OTHER     0x40

//
// Prefix Information option flags.
//
#define ND_PREFIX_FLAG_ON_LINK          0x80
#define ND_PREFIX_FLAG_AUTONOMOUS       0x40


//
// Stuff to handle in-kernel ping functionality.
//
typedef void (*EchoRtn)(void *, IP_STATUS, IPv6Addr *, void *, uint);

typedef struct EchoControl {
    struct EchoControl *Next;  // Next control structure in list.
    ulong TimeoutTimer;        // Timeout value (in IPv6Timer ticks).
    EchoRtn CompleteRoutine;   // Routine to call when completing request.
    ulong Seq;                 // Sequence number of this ping request.
    LARGE_INTEGER WhenIssued;  // Timestamp (in system timer ticks since boot).
    void *ReplyBuf;            // Buffer to store replies.
    ulong ReplyBufLen;         // Size of reply buffer.
} EchoControl;

IP_STATUS
ICMPv6EchoRequest(void *InputBuffer, uint InputBufferLength,
                  EchoControl *ControlBlock, EchoRtn Callback);

NTSTATUS
ICMPv6EchoComplete(EchoControl *ControlBlock,
                   IP_STATUS Status, IPv6Addr *Address,
                   void *Data, uint DataSize, ULONG *BytesReturned);

void
ICMPv6Send(
    NetTableEntry *NTE,   // NTE to send on (optional - NULL if don't care).
    PNDIS_PACKET Packet,  // Packet to send.
    uint IPv6Offset,      // Offset to IPv6 header in packet.
    uint ICMPv6Offset,    // Offset to ICMPv6 header in packet.
    IPv6Header *IP,       // Pointer to IPv6 header.
    uint PayloadLength,   // Length of IPv6 payload in bytes.
    ICMPv6Header *ICMP);  // Pointer to ICMPv6 header.

//
// An ICMPv6 error tuple
//
typedef struct ICMPv6MsgType {
    uchar Type;
    uchar Code;
    uint Value;            // holds "unused", "pointer", or "mtu" value
} ICMPv6MsgType;

int
ICMPv6Reply(
    NetTableEntry *RecvNTE,     // NTE packet was received on.
    IPv6Packet *Packet,         // Packet handed to us by link layer.
    ICMPv6MsgType *Type);       // Specifies msg type, code, and data pointer 

#endif  // ICMPv6_INCLUDED


