// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// IPSec extension headers and data structures.
// From the specifications approved by the IESG in mid-May 1998.
//

#ifndef IPSEC_INCLUDED
#define IPSEC_INCLUDED

// 
// Symbolic names of supported authentication algorithms.
//
#define IPSec_Auth_Dummy 0
#define IPSec_HMAC_MD5   1

//
// Encapsulating Security Payload header and trailer.
// The protocol number for ESP is 50.
//
// The trailer has 0-255 bytes of padding but needs to
// have the Pad Length and Next Header field end up aligned
// on the 32 bit boundary.
// TODOTODOTODO
// On send, after lookup in the Security Policy Database,
// packet construction should anticipate the padding need
// and make it easy to place the trailer.
//

typedef struct ESPHeader{
    ulong SPI;    // Security Parameters Index.
    ulong Seqno;  // Sequence number for anti-replay algorithms. 
} ESPHeader;

typedef struct ESPTrailer{
    ushort LastWord;        // Last 16 bits of pad or data.
    uchar PadLength;
    uchar NextHeader;
    PNDIS_BUFFER AuthData;  // Space to overlay specific algorithm fields.
                            // HMAC-MD5 uses 128 bits, total.
} ESPTrailer;


//
// Authentication Header structure.
// The protocol number for AH is 51.
//
typedef struct AHHeader {
   uchar NextHeader;
   uchar PayloadLen;
   ushort Reserved;
   ulong SPI;              // Security Parameters Index.
   ulong Seqno;            // Sequence number for anti-replay algorithms.
   PNDIS_BUFFER AuthData;  // Space to overlay specific algorithm fields.
                           // HMAC-MD5 uses 128 bits, total.
} AHHeader;
 

//
// Mandatory IPSec modes for hosts and security gateways, from
// the Security Architecture document section 4.5.
//
// REVIEWREVIEW
// N.B. We will not support general nestings of IPSEC.
//
// Host to Host Modes.
//
#define IPSec_AH_Transport            1
#define IPSec_ESP_Transport           2
#define IPSec_AH_ESP_Transport        3
#define IPSec_H2H_AH_Tunnel           4
#define IPSec_H2H_ESP_Tunnel          5
//
// Gateway to Gateway Modes.
//
// REVIEWREVIEW 
// We will not support these now because
// implementation of a security gateway
// is not IPv6 specific.
//
#define IPSec_G2G_AH_Tunnel           6
#define IPSec_G2G_ESP_Tunnel          7
#define IPSec_G2G_AH_ESP_Tunnel       8
#define IPSec_Types_MAX               9  
//
// Combined Host-Gateway Modes.
//
// REVIEWREVIEW
// We will also not support these.
// FYI the notation is that from left to right
// shows the inner header then the outer header.
//
// H2H_AH_Transport_G2G_ESP_Tunnel
// H2H_ESP_Transport_G2G_ESP_Tunnel
// H2H_ESP_Transport_G2G_AH_Tunnel
// H2H_AH_Transport_G2G_AH_Tunnel
//
// H2H_AH_Transport_H2G_ESP_Tunnel
// H2H_ESP_Transport_H2G_ESP_Tunnel
// H2H_ESP_Transport_H2G_AH_Tunnel
// H2H_AH_Transport_H2G_AH_Tunnel


//
// Security Association (SA) Database structure. 
//
// A unique SA is the tuple of the SPI, the destination address,
// and the IPSec protocol of the packet.  For packets with
// multiple IPSec extension headers, there are multiple SAs,
// comprising an SA Bundle.
//
typedef struct SecAssocEntry {           // a.k.a. SAE
    SecAssocEntry *SAEPrev;
    SecAssocEntry *SAENext;
    ulong SPI;                           // Security Parameter Index.
    IPv6Addr DestAdrr;                   // Destination address.
    ushort IPSecProto;                   // AH or ESP for authentication.
    ushort IPSecMode;                    // For use by sending side.
    ulong SASeqno;                       // Used by anti-replay algorithms.
    void (*IPSecAuth) (void *State, void *Data, uint Len);  
                                         // Pointer to the authentication
                                         // algorithm function to be used
                                         // in this security association.
                                         // We implement only the minimum
                                         // set so this function pointer is
                                         // simple.
} SecAssocEntry;


//
// Security Policy flag, one of which is associated with
// each match of a packet to the Security Policy Database.
// Following Security Architecture 4.4.1 paragraph 2.
//
#define SPF_NOSEC_BYPASSOK     0x1       // Matched packets are expected
                                         // to lack IPSec protocols; they
                                         // have an entry (usually wildcard)
                                         // with NULL SAList in the SPD.
                                         //
#define SPF_NOSEC_DISCARD      0x2       // Matched packets have 
                                         // mandatory use of their
                                         // entry's SAList, so if any
                                         // IPSec is missing, discard.
                                         // 
#define SPF_IPSEC_APPLY        0x4       // Security services are used.
                                         //
#define SPF_IPSEC_APPCHOICE    0x8       // Used by sending side, for
                                         // a policy permitting the security
                                         // service to be used on some
                                         // packets by application choice.
                                         // This is the sender-side of
                                         // a fine-granularity security
                                         // policy.

//
// Security Policy Database structure.
//
// N.B. We implement the mandatory selectors.
//
// Any traffic needs to be matched to a database entry.
// Organizing this database so that the match for traffic
// with a bypass policy is very fast would be desireable.
//
typedef struct SecPolicyDBEntry {        // a.k.a SPDE
    struct SecPolicyDBEntry *Next;
    struct SecPolicyDBEntry *Prev;
    uint SecPolicyFlag;                  // One of the three Security
                                         // Architecture sec. 4.4.1 policies
                                         // or application per-packet option
                                         // (see above).
                                         //
    uint SecPolicyMode;                  // One of the five supported
                                         // modes derived from the Security
                                         // Architecture sec. 4.5. (see above).
                                         //
    struct SecAssocEntry *SAList;        // Ordered list of the SAs for  
                                         // packets matching this SPDE.
                                         // May be NULL.
                                         //
    uint     DestAddrFunction;           // Ranged, masked or single.
    IPv6Addr DestAddr;                   // Cannot be NULL.
    IPv6Addr DestAddrData;               // Mask or end of range.
                                         //
    uint     SrcAddrFunction;            // Ranged, masked or single.
    IPv6Addr SrcAddr;                    // Cannot be NULL.
    IPv6Addr SrcAddrData;                // Mask or end of range.
                                         //
    ushort   TransportProto;             // If NULL, protocol is opaque (and 
                                         // the ports should be skipped too).
                                         //
    ushort   DestPortFunction;           // Ranged, masked or single.
    ushort   DestPort;                   // If NULL, both ports are opaque.
    ushort   DestPortData;               // Mask or end of range.
                                         // 
    ushort   SrcPortFunction;            // Ranged, masked or single.
    ushort   SrcPort;                    // 
    ushort   SrcPortData;                // Mask or end of range.
                                         // 
    PNDIS_BUFFER Name;                   // Required selector type for
                                         // system or user@system
                                         // identifiers.  N.B. This can't
                                         // be a string, because the X.500
                                         // name types can be non-ascii
                                         // character sets.
                                         // If NULL, this selector is
                                         // opaque.  
} SecurityPolicyDBEntry;

#endif  // IPSEC_INCLUDED
