// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Transmit routines for Internet Protocol Version 6.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6opts.h"
#include "ip6def.h"
#include "route.h"
#include "neighbor.h"
#include "fragment.h"

static ulong FragmentId = 0;


//* IPv6FreePacket - free an IPv6 packet.
//
//  Frees a packet whose buffers were allocated from the IPv6BufferPool.
//
void
IPv6FreePacket(PNDIS_PACKET Packet)
{
    PNDIS_BUFFER Buffer, NextBuffer;

    //
    // Free all the buffers in the packet.
    // Start with the first buffer in the packet and follow the chain.
    //
    NdisQueryPacket(Packet, NULL, NULL, &Buffer, NULL);
    for (; Buffer != NULL; Buffer = NextBuffer) {
        VOID *Mem;
        UINT Unused;

        //
        // Free the buffer descriptor back to IPv6BufferPool and its
        // associated memory back to the heap.  Not clear if it would be
        // safe to free the memory before the buffer (because the buffer
        // references the memory), but this order should definitely be safe.
        //
        NdisGetNextBuffer(Buffer, &NextBuffer);
        NdisQueryBuffer(Buffer, &Mem, &Unused);
        NdisFreeBuffer(Buffer);
        ExFreePool(Mem);
    }

    //
    // Free the packet back to IPv6PacketPool.
    //
    NdisFreePacket(Packet);
}


//* IPv6SendComplete - IP send complete handler.
//
//  Called by the link layer when a send completes.  We're given a pointer to
//  a net structure, as well as the completing send packet and the final status
//  of the send.
//
void                      //  Returns: Nothing.
IPv6SendComplete(
    void *Context,        // Context we gave to the link layer on registration.
    PNDIS_PACKET Packet,  // Packet completing send.
    NDIS_STATUS Status)   // Final status of send.
{
    KIRQL OldIrql;         // For locking the multicast link-layer address list.
    UNREFERENCED_PARAMETER(Context);
    UNREFERENCED_PARAMETER(Status);

    //
    // The packet should have a completion handler specified for us to call.
    //
    ASSERT(PC(Packet)->CompletionHandler != NULL);

    //
    // Call the packet's designated completion handler.
    //
    (*PC(Packet)->CompletionHandler)(Packet);
}


//* IPv6Send0 - Send a fully-formed IPv6 packet.
//
//  Lowest-level send routine.  We already know the first-hop destination and
//  have a completed packet ready to send.  All we really do here
//  is check & update the NCE's neighbor discovery state.
//
//  Discovery Address is the source address to use in neighbor 
//  discovery solicitations.
//
//  (This is inlined here instead of being a function in neighbor.c.)
//
//  Callable from thread or DPC context.
//
void
IPv6Send0(
    PNDIS_PACKET Packet,        // Packet to send.
    uint Offset,                // Offset from start of Packet to IP header.
    NeighborCacheEntry *NCE,    // First-hop neighbor information.
    IPv6Addr *DiscoveryAddress) //Address to use for neighbor discovery.
{
    KIRQL OldIrql1;     // For locking the interface.
    Interface *IF;      // Interface to send via.
    LARGE_INTEGER Now;  // Time since boot in system interval timer ticks.

    ASSERT(NCE != NULL);
    IF = NCE->IF;

    KeAcquireSpinLock(&IF->Lock, &OldIrql1);

    //
    // Check the Neighbor Discovery Protocol state of our Neighbor to
    // insure that we have current information to work with.  We don't
    // have a timer going off to drive this in the common case, but
    // instead check the reachability timestamp directly here.
    //
    switch (NCE->NDState) {
    case ND_STATE_PERMANENT:
        //
        // This neighbor is always valid.
        //
        break;

    case ND_STATE_REACHABLE:
        //
        // Common case.  We've verified neighbor reachability within
        // the last 'ReachableTime' ticks of the system interval timer.
        // If the time limit hasn't expired, we're free to go.
        //
        // Note that the following arithmetic will correctly handle wraps
        // of the system interval timer's tick counter (however unlikely it
        // is that a machine will stay up that long without rusting away).
        //
        KeQueryTickCount(&Now);
        if ((LONGLONG)(NCE->LastReachable.QuadPart + IF->ReachableTime) -
            Now.QuadPart > 0) {
            //
            // Got here within the time limit.  Just send it.
            //
            break;
        }

        //
        // Too long since last send.  Entry went stale.  Conceptually,
        // we've been in the STALE state since the above quantity went
        // positive.  So just drop on into it now...
        // 

    case ND_STATE_STALE:
        //
        // We have a stale entry in our neighbor cache.  Go into DELAY
        // state, start the delay timer, and send the packet anyway.
        //
        NCE->NDState = ND_STATE_DELAY;
        NCE->EventTimer = DELAY_FIRST_PROBE_TIME;
        break;

    case ND_STATE_DELAY:
    case ND_STATE_PROBE:
        //
        // While in the DELAY and PROBE states, we continue to send to our
        // cached address and hope for the best.  Sending more packets in
        // either state has no other effect.
        //
        break;

    case ND_STATE_INCOMPLETE: {
        PNDIS_PACKET OldPacket;
        int SendSolicit;

        //
        // We do not have a valid link-level address for the neighbor.
        // We must queue the packet, pending neighbor discovery.
        // Remember the packet's offset in the Packet6Context area.
        // BUGBUG: For now, wait queue is just one packet deep.
        //
        OldPacket = NCE->WaitQueue;
        PC(Packet)->pc_offset = Offset;
        PC(Packet)->DiscoveryAddress = *DiscoveryAddress;
        NCE->WaitQueue = Packet;

        //
        // If we have not started neighbor discovery yet,
        // do so now by sending the first solicit.
        //
        if (SendSolicit = (NCE->EventCount == 0)) {

            NCE->EventCount = 1;
            NCE->EventTimer = IF->RetransTimer;
        }

        KeReleaseSpinLock(&IF->Lock, OldIrql1);

        if (SendSolicit) {
            NeighborSolicitSend(NCE, DiscoveryAddress);
        }

        if (OldPacket != NULL) {
            IPv6SendComplete(NULL, OldPacket, NDIS_STATUS_REQUEST_ABORTED);
        }
        return;
    }

    default:
        //
        // Should never happen.
        //
        ASSERTMSG("IPv6Send0: Invalid Neighbor Cache NDState field!\n", FALSE);
    }

    //
    // Move the NCE to the head of the LRU list,
    // because we are using it to send a packet.
    //
    if (NCE != IF->FirstNCE) {
        //
        // Remove NCE from the list.
        //
        NCE->Next->Prev = NCE->Prev;
        NCE->Prev->Next = NCE->Next;

        //
        // Add NCE to the head of the list.
        //
        NCE->Next = IF->FirstNCE;
        NCE->Next->Prev = NCE;
        NCE->Prev = SentinelNCE(IF);
        NCE->Prev->Next = NCE;
    }


    //
    // Unlock before transmitting the packet.
    // This means that there is a very small chance that NCE->LinkAddress
    // could change out from underneath us. (For example, if we process
    // an advertisement changing the link-level address.)
    // In practice this won't happen, and if it does the worst that
    // will happen is that we'll send a packet somewhere strange.
    // The alternatives are copying the LinkAddress or holding the lock.
    //
    KeReleaseSpinLock(&IF->Lock, OldIrql1);

    (*IF->Transmit)(IF->LinkContext, Packet, Offset, NCE->LinkAddress);
}


//* IPv6SendFragments - Fragment an IPv6 datagram.
//
//  Helper routine for creating and sending IPv6 fragments.
//  Called from IPv6Send1 when the datagram is bigger than the path MTU.
//
//  The PathMTU is passed separately so that we use a consistent value.
//  The value in the RCE is subject to change.
//
//  REVIEW: The algorithm used here generates all the fragments at once.
//  It might be better to generate the fragments one at a time,
//  waiting for each send to complete before moving on.
//  This would mean breaking up the work and having a special
//  send-completion routine.
//
void
IPv6SendFragments(
    PNDIS_PACKET Packet,   // Packet to send.
    uint Offset,           // Offset from start of Packet to IP header.
    IPv6Header *IP,        // Pointer to Packet's IPv6 header.
    uint PayloadLength,    // Packet payload length.
    RouteCacheEntry *RCE,  // First-hop neighbor information.
    uint PathMTU)          // PathMTU to use when fragmenting.
{
    NeighborCacheEntry *NCE = RCE->NCE;
    NDIS_STATUS Status;
    PNDIS_PACKET FragPacket;
    PNDIS_BUFFER FragBuffer;
    FragmentHeader FragHdr;
    uchar *Mem;
    uint MemLen;
    uint PktOffset;
    uint UnfragBytes;
    uint BytesLeft;
    uint BytesSent;
    uchar HdrType;
    uchar *tbuf;
    PNDIS_BUFFER SrcBuffer;
    uint SrcOffset;
    uint NextHeaderOffset;
    uint FragPayloadLength;

    //
    // Determine the 'unfragmentable' portion of this packet.
    // We do this by scanning through all extension headers,
    // and noting the last occurrence, if any, of
    // a routing or hop-by-hop header.
    // We do not assume the extension headers are in recommended order,
    // but otherwise we assume that the headers are well-formed.
    // BUGBUG: We assume that they are contiguous.
    //
    UnfragBytes = sizeof *IP;
    HdrType = IP->NextHeader;
    NextHeaderOffset = (uchar *)&IP->NextHeader - (uchar *)IP;
    tbuf = (uchar *)(IP + 1);
    while ((HdrType == IP_PROTOCOL_HOP_BY_HOP) ||
           (HdrType == IP_PROTOCOL_ROUTING) ||
           (HdrType == IP_PROTOCOL_DEST_OPTS)) {
        ExtensionHeader *EHdr = (ExtensionHeader *) tbuf;
        uint EHdrLen = (EHdr->HeaderExtLength + 1) * 8;

        tbuf += EHdrLen;
        if (HdrType != IP_PROTOCOL_DEST_OPTS) {
            UnfragBytes = tbuf - (uchar *)IP;
            NextHeaderOffset = (uchar *)&EHdr->NextHeader - (uchar *)IP;
        }
        HdrType = EHdr->NextHeader;
    }

    //
    // Check that we can actually fragment this packet.
    // If the unfragmentable part is too large, we can't.
    // We need to send at least 8 bytes of fragmentable data
    // in each fragment.
    //
    if (UnfragBytes + FRAG_HDR_LEN + 8 > PathMTU) {
        DbgPrint("IPv6SendFragments: can't fragment\n");
        Status = NDIS_STATUS_REQUEST_ABORTED;
        goto ErrorExit;
    }

    FragHdr.NextHeader = HdrType;
    FragHdr.Reserved = 0;
    FragHdr.Id = net_long(InterlockedIncrement(&FragmentId));

    //
    // Initialize SrcBuffer and SrcOffset, which point
    // to the fragmentable data in the packet.
    // SrcOffset is the offset into SrcBuffer's data,
    // NOT an offset into the packet.
    //
    NdisQueryPacket(Packet, NULL, NULL, &SrcBuffer, NULL);
    SrcOffset = Offset + UnfragBytes;

    //
    // Create new packets of MTU size until all data is sent.
    //
    BytesLeft = sizeof *IP + PayloadLength - UnfragBytes;
    PktOffset = 0; // relative to fragmentable part of original packet

    while (BytesLeft != 0) {
        //
        // Allocate packet (and a buffer) and Memory for new fragment
        //
        NdisAllocatePacket(&Status, &FragPacket, IPv6PacketPool);
        if (Status != NDIS_STATUS_SUCCESS) {
            goto ErrorExit;
        }
        PC(FragPacket)->CompletionHandler = IPv6FreePacket;

        //
        // Determine new IP payload length (a multiple of 8) and 
        // and set the Fragment Header offset.
        //
        if ((BytesLeft + UnfragBytes + FRAG_HDR_LEN) > PathMTU) {
            BytesSent = (PathMTU - UnfragBytes - FRAG_HDR_LEN) &~ 7;
            // Not the last fragment, so turn on the M bit.
            FragHdr.OffsetFlag = net_short((ushort)(PktOffset | 1));
        } else {
            BytesSent = BytesLeft;
            FragHdr.OffsetFlag = net_short((ushort)PktOffset);
        }

        MemLen = Offset + UnfragBytes + FRAG_HDR_LEN + BytesSent;
        Mem = ExAllocatePool(NonPagedPool, MemLen);
        if (Mem == NULL) {
            NdisFreePacket(FragPacket);
            Status = NDIS_STATUS_RESOURCES;
            goto ErrorExit;
        }

        //
        // Allocate a buffer head and put it on the packet chain.
        //
        NdisAllocateBuffer(&Status, &FragBuffer, IPv6BufferPool, Mem, MemLen);
        if (Status != NDIS_STATUS_SUCCESS) {
            ExFreePool(Mem);
            NdisFreePacket(FragPacket);
            goto ErrorExit;
        }
        NdisChainBufferAtFront(FragPacket, FragBuffer);

        //
        // Copy IP header, Frag Header, and a portion of data to fragment.
        //
        RtlCopyMemory(Mem + Offset, IP, UnfragBytes);
        RtlCopyMemory(Mem + Offset + UnfragBytes, &FragHdr, FRAG_HDR_LEN);
        CopyFromBufferChain(Mem + Offset + UnfragBytes + FRAG_HDR_LEN,
                            SrcBuffer, SrcOffset, BytesSent,
                            &SrcBuffer, &SrcOffset);

        //
        // Correct the PayloadLength and NextHeader fields.
        //
        FragPayloadLength = UnfragBytes + FRAG_HDR_LEN + BytesSent -
                                IPv6_HDR_LEN;
        ASSERT(FragPayloadLength <= MAX_IPv6_PAYLOAD);
        ((IPv6Header *)(Mem + Offset))->PayloadLength =
            net_short((ushort) FragPayloadLength);
        ASSERT(Mem[Offset + NextHeaderOffset] == HdrType);
        Mem[Offset + NextHeaderOffset] = IP_PROTOCOL_FRAGMENT;

        BytesLeft -= BytesSent;
        PktOffset += BytesSent;

        //
        // Send the fragment.
        //
        IPv6Send0(FragPacket, Offset, NCE, &(RCE->NTE->Address));
    }

    //
    // BUGBUG: Really we should be monitoring send-completion
    // status information for each of the fragments.
    //
    Status = NDIS_STATUS_SUCCESS;
  ErrorExit:
    IPv6SendComplete(NULL, Packet, Status);
}


//* IPv6Send1 - Send an IPv6 datagram.
//
//  Slightly higher-level send routine.  We have a completed datagram and a
//  RCE indicating where to direct it to.  Here we deal with any packetization
//  issues (inserting a Jumbo Payload option, fragmentation, etc.) that are
//  necessary, and pick a NCE for the first hop.
//
void
IPv6Send1(
    PNDIS_PACKET Packet,   // Packet to send.
    uint Offset,           // Offset from start of Packet to IP header.
    IPv6Header *IP,        // Pointer to Packet's IPv6 header.
    uint PayloadLength,    // Packet payload length.
    RouteCacheEntry *RCE)  // First-hop neighbor information.
{
    uint PacketLength;        // Size of complete IP packet in bytes.
    NeighborCacheEntry *NCE;  // First-hop neighbor information.
    uint PathMTU;

    //
    // See if we need to insert a Jumbo Payload option.
    //
    if (PayloadLength > MAX_IPv6_PAYLOAD) {
        // BUGBUG: Add code to insert a Jumbo Payload hop-by-hop option here.
        DbgPrint("IPv6Send1: attempted to send a Jumbo Payload!\n");
        IPv6SendComplete(NULL, Packet, NDIS_STATUS_REQUEST_ABORTED);
        return;
    }

    //
    // Check the path's MTU.  If we're larger, fragment.
    //
    PacketLength = PayloadLength + sizeof(IPv6Header);
    if (PacketLength > (PathMTU = RCE->PathMTU)) {

        IPv6SendFragments(Packet, Offset, IP, PayloadLength, RCE, PathMTU);
    }
    else {
        //
        // Fill in packet's PayloadLength field.
        //
        IP->PayloadLength = net_short((ushort)PayloadLength);

        //
        // We only have one NCE per RCE for now,
        // so picking one is really easy...
        //
        NCE = RCE->NCE;

        IPv6Send0(Packet, Offset, NCE, &(RCE->NTE->Address));
    }
}
