// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Various subroutines for Internet Protocol Version 6 
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "fragment.h"
#include "neighbor.h"
#include "route.h"

extern ulong RandomValue;

//* Random - Generate a psuedo random value between 1 and 2^32.
//
//  This routine is a quick and dirty psuedo random number generator.
//  It has the advantages of being fast and consuming very little
//  memory (for either code or data).  The random numbers it produces are
//  not of the best quality, however.  A much better generator could be
//  had if we were willing to use an extra 256 bytes of memory for data.
//
//  This routine uses the linear congruential method (see Knuth, Vol II),
//  with specific values for the multiplier and constant taken from
//  Numerical Recipes in C Second Edition by Press, et. al.
//
ulong  // Returns: A random value between 1 and 2^32.
Random(void)
{
    //
    // The algorithm is R = (aR + c) mod m, where R is the random number,
    // a is a magic multiplier, c is a constant, and the modulus m is the
    // maximum number of elements in the period.  We chose our m to be 2^32
    // in order to get the mod operation for free.
    // BUGBUG: What about concurrent calls?
    //
    RandomValue = (1664525 * RandomValue) + 1013904223;

    return RandomValue;
}


//* CopyFromBufferChain
//
//  Copy from a buffer chain to a flat buffer.
//  Returns the Next buffer/offset, for subsequent calls.
//
void
CopyFromBufferChain(
    void *DstData,
    PNDIS_BUFFER SrcBuffer,
    uint SrcOffset,
    uint Length,
    PNDIS_BUFFER *NextBuffer,
    uint *NextOffset)
{
    void *SrcData;
    uint SrcSize;
    uint Bytes;

    for (;;) {
        NdisQueryBuffer(SrcBuffer, &SrcData, &SrcSize);
        ASSERT(SrcOffset <= SrcSize);
        Bytes = SrcSize - SrcOffset;
        if (Bytes > Length)
            Bytes = Length;

        RtlCopyMemory(DstData, (uchar *)SrcData + SrcOffset, Bytes);

        (uchar *)DstData += Bytes;
        SrcOffset += Bytes;
        Length -= Bytes;

        if (Length == 0)
            break;

        NdisGetNextBuffer(SrcBuffer, &SrcBuffer);
        SrcOffset = 0;
    }

    *NextBuffer = SrcBuffer;
    *NextOffset = SrcOffset;
}


//
// Checksum support.
// On NT, there are architecture-specific assembly routines for the core
// calculation.
//

//* ChecksumPacket - Calculate the Internet checksum of a packet.
//
//  Calculates the checksum of packet data. The data may be supplied
//  either with the Packet/Offset arguments, or (if Packet is NULL)
//  the Data argument. In either case, Length specifies how much
//  data to checksum.
//
//  The Packet is assumed to contain (at least) Offset + Length bytes.
//
//  Also calculates and adds-in the pseudo-header checksum,
//  using IP->Source, IP->Dest, Length, and NextHeader.
//  (With options, we may have IP->NextHeader != NextHeader.)
//
ushort
ChecksumPacket(
    PNDIS_PACKET Packet,  // Packet with data to checksum.
    uint Offset,          // Offset into packet where data starts.
    uchar *Data,          // If Packet is NULL, data to checksum.
    uint Length,          // Length of packet data.
    IPv6Header *IP,       // Header with Source & Dest addresses.
    uchar NextHeader)     // Protocol type for pseudo-header.
{
    PNDIS_BUFFER Buffer;
    uint Checksum;
    uint PayloadLength;
    uint Size;
    uint TotalSummed;

    //
    // Start with the pseudo-header.
    // NB: IP->Dest follows IP->Source with no gap.
    //
    Checksum = Cksum(&IP->Source, sizeof IP->Source + sizeof IP->Dest);
    PayloadLength = net_long(Length);
    Checksum += (PayloadLength >> 16) + (PayloadLength & 0xffff);
    Checksum += (NextHeader << 8);
    
    if (Packet == NULL) {
        //
        // We do not have to initialize Buffer.
        // The checksum loop below will exit before trying to use it.
        //
        Size = Length;
    } else {
        UINT Unused;

        //
        // Skip over Offset bytes in the packet.
        //

        NdisGetFirstBufferFromPacket(Packet, &Buffer, &Data, &Size, &Unused);
        for (;;) {

            //
            // There is a boundary case here: the Packet contains
            // exactly Offset bytes total, and Length is zero.
            // Checking Offset <= Size instead of Offset < Size
            // makes this work.
            //
            if (Offset <= Size) {
                Data += Offset;
                Size -= Offset;
                break;
            }

            Offset -= Size;
            NdisGetNextBuffer(Buffer, &Buffer);
            NdisQueryBuffer(Buffer, &Data, &Size);
        }
    }
    for (TotalSummed = 0;;) {
        ushort Temp;

        //
        // Size might be bigger than we need,
        // if there is "extra" data in the packet.
        //
        if (Size > Length)
            Size = Length;

        Temp = Cksum(Data, Size);
        if (TotalSummed & 1) {
            // We're at an odd offset into the logical buffer,
            // so we need to swap the bytes that Cksum returns.
            Checksum += (Temp >> 8) + ((Temp & 0xff) << 8);
        } else {
            Checksum += Temp;
        }
        TotalSummed += Size;

        Length -= Size;
        if (Length == 0)
            break;
        NdisGetNextBuffer(Buffer, &Buffer);
        NdisQueryBuffer(Buffer, &Data, &Size);
    }

    //
    // Wrap in the carries to reduce Checksum to 16 bits.
    // (Twice is sufficient because it can only overflow once.)
    //
    Checksum = (Checksum >> 16) + (Checksum & 0xffff);
    Checksum += (Checksum >> 16);

    //
    // Take ones-complement and replace 0 with 0xffff.
    //
    Checksum = (ushort) ~Checksum;
    if (Checksum == 0)
        Checksum = 0xffff;

    return (ushort) Checksum;
}

//* ConvertSecondsToTicks
//
//  Convert seconds to timer ticks.
//  A value of DAD_LIFETIME_INFINITE (0xffffffff) indicates infinity.
//
uint
ConvertSecondsToTicks(uint Seconds)
{
    uint Ticks;

    //
    // BUGBUG: Overflow?
    //
    if (Seconds == DAD_LIFETIME_INFINITE)
        Ticks = DAD_LIFETIME_INFINITE;
    else
        Ticks = IPv6TimerTicks(Seconds);

    return Ticks;
}

//* ConvertMillisToTicks
//
//  Convert milliseconds to timer ticks.
//
uint
ConvertMillisToTicks(uint Millis)
{
    uint Ticks;

    //
    // BUGBUG: Overflow/underflow?
    //
    Ticks = IPv6TimerTicks(Millis) / 1000;
    if (Ticks == 0 && Millis != 0)
        Ticks = 1;

    return Ticks;
}

//* IPv6Timeout - Perform various housekeeping duties periodically.
//
//  Neighbor discovery, fragment reassembly, ICMP ping, etc. all have
//  time-dependent parts.  Check for timer expiration here.
//
void
IPv6Timeout(
    PKDPC MyDpcObject,  // The DPC object describing this routine.
    void *Context,      // The argument we asked to be called with.
    void *Unused1,
    void *Unused2)
{
    static uint RecalcReachableTimer = 0;
    int RecalcReachable;
    Interface *IF;

    UNREFERENCED_PARAMETER(Context);
    UNREFERENCED_PARAMETER(Unused1);
    UNREFERENCED_PARAMETER(Unused2);

    //
    // Recalculate ReachableTime every few hours,
    // even if no Router Advertisements are received.
    //
    if (RecalcReachableTimer == 0) {
        RecalcReachable = TRUE;
        RecalcReachableTimer = RECALC_REACHABLE_INTERVAL;
    } else {
        RecalcReachable = FALSE;
        RecalcReachableTimer--;
    }

    //
    // Handle per-interface timeouts.
    // BUGBUG - The interface list should be locked.
    //
    for (IF = IFList; IF != NULL; IF = IF->Next) {
        //
        // Handle per-neighbor timeouts.
        //
        NeighborCacheTimeout(IF);

        //
        // Handle router solicitations.
        // The timer check here is only an optimization,
        // because it is made without holding the appropriate lock.
        //
        if (IF->RSTimer != 0)
            RouterSolicitTimeout(IF);

        //
        // Recalculate the reachable time.
        //
        if (RecalcReachable) {

            KeAcquireSpinLockAtDpcLevel(&IF->Lock);
            IF->ReachableTime = CalcReachableTime(IF->BaseReachableTime);
            KeReleaseSpinLockFromDpcLevel(&IF->Lock);
        }
    }

    //
    // Handle per-NTE timeouts.
    //
    NetTableTimeout();

    //
    // Handle prefix-list timeouts.
    //
    PrefixListTimeout();

    //
    // Handle default router timeouts.
    //
    RouterListTimeout();

    //
    // If there's a possibility we have one or more outstanding echo requests,
    // call out to ICMPv6 to handle them.  Note that since we don't grab the
    // lock here, there may be none by the time we get there.  This just saves
    // us from always having to call out.
    //
    if (ICMPv6OutstandingEchos != NULL) {
        //
        // Echo requests outstanding.
        //
        ICMPv6EchoTimeout();
    }

    //
    // Check for timed-out fragments.
    //
    if (ReassemblyList != NULL) {
        ReassemblyTimeout();
    }

    // 
    // Process all multicast groups with timers running.  Timers are used to
    // response to membership queries sent to us by the first-hop router.
    //
    if (QueryList != NULL)
        MLDTimeout();

}

//* AdjustPacketBuffer
//
//  Takes an NDIS Packet that has some spare bytes available
//  at the beginning and adjusts the size of that available space.
//
//  When we allocate packets, we often do not know a priori on which
//  link the packets will go out.  However it is much more efficient
//  to allocate space for the link-level header along with the rest
//  of the packet.  Hence we leave space for the maximum link-level header,
//  and each individual link layer uses AdjustPacketBuffer to shrink
//  that space to the size that it really needs.
//
//  AdjustPacketBuffer is needed because the sending calls (in both
//  the NDIS and TDI interfaces) do not allow the caller to specify
//  an offset of data to skip over.
//
//  Note that this code is NT-specific, because it knows about the
//  internal fields of NDIS_BUFFER structures.
//
void *
AdjustPacketBuffer(
    PNDIS_PACKET Packet,  // Packet to adjust.
    uint SpaceAvailable,  // Extra space available at start of first buffer.
    uint SpaceNeeded)     // Amount of space we need for the header.
{
    PMDL Buffer;
    uint Adjust;

    // Get first buffer on packet chain.
    NdisQueryPacket(Packet, NULL, NULL, &Buffer, NULL);

    //
    // The remaining space in the packet should all be in the first buffer.
    //
    ASSERT(SpaceAvailable <= Buffer->ByteCount);

    Adjust = SpaceAvailable - SpaceNeeded;
    if (Adjust == 0) {
        //
        // There is exactly the right amount of space left.
        // This is the common case.
        //
    } else if ((int)Adjust > 0) {
        //
        // There is too much space left.
        // Because NdisSend doesn't have an Offset argument,
        // we need to temporarily "shrink" the buffer.
        //
        (uchar *)Buffer->MappedSystemVa += Adjust;
        Buffer->ByteCount -= Adjust;
        Buffer->ByteOffset += Adjust;

        if (Buffer->ByteOffset >= PAGE_SIZE) {
            ULONG FirstPage;

            //
            // Need to "remove" the first physical page
            // by shifting the array up one page.
            // Save it at the end of the array.
            //
            FirstPage = ((PULONG)(Buffer + 1))[0];
            RtlMoveMemory(&((PULONG)(Buffer + 1))[0],
                          &((PULONG)(Buffer + 1))[1],
                          Buffer->Size - sizeof *Buffer - sizeof(ULONG));
            ((PULONG)((uchar *)Buffer + Buffer->Size))[-1] = FirstPage;

            (uchar *)Buffer->StartVa += PAGE_SIZE;
            Buffer->ByteOffset -= PAGE_SIZE;
        }
    } else { // Adjust < 0
        //
        // Not enough space.
        // Shouldn't happen in the normal send path.
        // REVIEW: This is a potential problem when forwarding packets
        // from an interface with a short link-level header
        // to an interface with a longer link-level header.
        // Should the forwarding code take care of this?
        //
        ASSERTMSG("AdjustPacketBuffer: Adjust < 0", FALSE);
    }

    //
    // Save away the adjustment for the completion callback,
    // which needs to undo our work with UndoAdjustPacketBuffer.
    //
    PC(Packet)->pc_offset = Adjust;

    //
    // Return a pointer to the buffer.
    //
    return Buffer->MappedSystemVa;
}

//* UndoAdjustPacketBuffer
//
//  Undo the effects of AdjustPacketBuffer.
//
//  Note that this code is NT-specific, because it knows about the
//  internal fields of NDIS_BUFFER structures.
//
void
UndoAdjustPacketBuffer(
    PNDIS_PACKET Packet)  // Packet we may or may not have previously adjusted.
{
    uint Adjust;

    Adjust = PC(Packet)->pc_offset;
    if (Adjust != 0) {
        PMDL Buffer;

        //
        // We need to undo the adjustment made in AdjustPacketBuffer.
        // This may including shifting the array of page info.
        //

        // Get first buffer on packet chain.
        NdisQueryPacket(Packet, NULL, NULL, &Buffer, NULL);

        if (Buffer->ByteOffset < Adjust) {
            ULONG FirstPage;

            (uchar *)Buffer->StartVa -= PAGE_SIZE;
            Buffer->ByteOffset += PAGE_SIZE;

            FirstPage = ((PULONG)((uchar *)Buffer + Buffer->Size))[-1];
            RtlMoveMemory(&((PULONG)(Buffer + 1))[1],
                          &((PULONG)(Buffer + 1))[0],
                          Buffer->Size - sizeof *Buffer - sizeof(ULONG));
            ((PULONG)(Buffer + 1))[0] = FirstPage;
        }

        (uchar *)Buffer->MappedSystemVa -= Adjust;
        Buffer->ByteCount += Adjust;
        Buffer->ByteOffset -= Adjust;
    }
}

//* IsV4Compatible
//
//  Is this a v4-compatible address?
//
int
IsV4Compatible(IPv6Addr *Addr)
{
    return ((Addr->u.DWord[0] == 0) &&
            (Addr->u.DWord[1] == 0) &&
            (Addr->u.DWord[2] == 0) &&
            (Addr->u.DWord[3] > 1));
}

//* HasPrefix - Does an address have the given prefix?
//
int
HasPrefix(IPv6Addr *Addr, IPv6Addr *Prefix, uint PrefixLength)
{
    uchar *AddrBytes = Addr->u.Byte;
    uchar *PrefixBytes = Prefix->u.Byte;

    //
    // Check that initial integral bytes match.
    //
    while (PrefixLength > 8) {
        if (*AddrBytes++ != *PrefixBytes++)
            return FALSE;
        PrefixLength -= 8;
    }

    //
    // Check any remaining bits.
    // Note that if PrefixLength is zero now, we should not
    // dereference AddrBytes/PrefixBytes.
    //
    if ((PrefixLength > 0) &&
        ((*AddrBytes >> (8 - PrefixLength)) !=
         (*PrefixBytes >> (8 - PrefixLength))))
        return FALSE;

    return TRUE;
}

//* CommonPrefixLength
//
//  Calculate the length of the longest prefix common
//  to the two addresses.
//
uint
CommonPrefixLength(IPv6Addr *Addr, IPv6Addr *Addr2)
{
    int i, j;

    //
    // Find first non-matching byte.
    //
    for (i = 0; ; i++) {
        if (i == sizeof(IPv6Addr))
            return 8 * i;

        if (Addr->u.Byte[i] != Addr2->u.Byte[i])
            break;
    }

    //
    // Find first non-matching bit (there must be one).
    //
    for (j = 0; ; j++) {
        uint Mask = 1 << (7 - j);

        if ((Addr->u.Byte[i] & Mask) != (Addr2->u.Byte[i] & Mask))
            break;
    }

    return 8 * i + j;
}

//* GetIPv6Header
//
//  Returns a pointer to the IPv6 header in a packet.
//  The header must be contiguous and in the first buffer.
//
IPv6Header UNALIGNED *
GetIPv6Header(PNDIS_PACKET Packet, uint Offset)
{
    PNDIS_BUFFER Buffer;
    VOID *VirtAddr;
    UINT BufLen, TotLen;

    NdisGetFirstBufferFromPacket(Packet, &Buffer, &VirtAddr, &BufLen, &TotLen);

    if (BufLen < Offset + sizeof(IPv6Header))
        return NULL;  // Header isn't in the first buffer.

    return (IPv6Header UNALIGNED *)((uchar *)VirtAddr + Offset);
}

//* PacketPullup - extend contiguous data region.
//
//  Pulls up more data from secondary packet buffers to create a contiguous
//  buffer of at least the given amount.
//
//  If a zero size contiguous region is requested, we always pullup the
//  entire next buffer.
//
uint  // Returns: new contiguous amount, or 0 if unable to satisfy request.
PacketPullup(
    IPv6Packet *Packet,  // Packet to pullup.
    uint Needed)         // Minimum amount of contiguous data to return.
{
    PNDIS_BUFFER Buffer;
    void *VirtAddr;
    uint BufLen, TotLen, Offset;

    //
    // REVIEW: Callers should check this themselves before calling?
    //
    if (Needed && Needed < Packet->ContigSize)
        return Packet->ContigSize;
    if (Needed > Packet->TotalSize)
        return 0;

    //
    // Check to see if we have any more buffers in this packet.
    // If not, we can't do a pullup.
    //
    if (Packet->NdisPacket == NULL)
        return 0;

    //
    // Run down secondary buffer chain until we have reached
    // the buffer corresponding to the current position.
    //
    NdisGetFirstBufferFromPacket(Packet->NdisPacket, &Buffer, &VirtAddr,
                                 &Offset, &TotLen);
    ASSERT(Packet->Position <= TotLen);
    while (Offset < Packet->Position) {
        NdisGetNextBuffer(Buffer, &Buffer);
        NdisQueryBuffer(Buffer, &VirtAddr, &BufLen);
        Offset += BufLen;
    }

    //
    // If our current contiguous region is the result of a previous pullup,
    // we'll have an active auxiliary data area.  Free it now and fix up
    // the various packet fields appropriately.
    //
    if (Packet->Auxiliary != NULL) {
        ExFreePool(Packet->Auxiliary);
        Packet->ContigSize = Offset - Packet->Position;
        Packet->Data = (uchar *)VirtAddr + (BufLen - Packet->ContigSize);
    }

    ASSERT(Packet->ContigSize == Offset - Packet->Position);
    ASSERT((uchar *)VirtAddr + (BufLen - Packet->ContigSize) == Packet->Data);

    //
    // Now get the buffer after the one containing the current position.
    //
    NdisGetNextBuffer(Buffer, &Buffer);
    NdisQueryBuffer(Buffer, &VirtAddr, &BufLen);

    //
    // In the case where there is no data remaining from the previous buffer
    // of interest (i.e. we've completely exhausted the previous contiguous
    // region), and this buffer is big enough to satisfy the request, then
    // we only need to update some values and we're done.
    //
    if (Packet->ContigSize == 0 && BufLen >= Needed) {
        Packet->Data = VirtAddr;
        Packet->ContigSize = BufLen;
        return Packet->ContigSize;
    }

    //
    // If this buffer is big enough to satisfy the request, then we'll
    // return a contiguous region containing the remainder of the previous
    // buffer (from current position onward) plus all of this buffer.
    //
    if (Packet->ContigSize + BufLen >= Needed) {
        Packet->Auxiliary = ExAllocatePool(NonPagedPool,
                                           Packet->ContigSize + BufLen);
        RtlCopyMemory(Packet->Auxiliary, Packet->Data, Packet->ContigSize);
        RtlCopyMemory((uchar *)Packet->Auxiliary + Packet->ContigSize,
                      VirtAddr, BufLen);
        Packet->ContigSize += BufLen;
    } else {
        uint CopyNow;

        //
        // We're going to need to run through at least another buffer to
        // satisfy this request.  So we only have to run down the chain
        // once, we'll limit the pullup to exactly the requested amount.
        // REVIEW: Alternatively, we could pullup the entire packet.
        //
        Packet->Auxiliary = ExAllocatePool(NonPagedPool, Needed);
        RtlCopyMemory(Packet->Auxiliary, Packet->Data, Packet->ContigSize);
        Needed -= Packet->ContigSize;
        Offset = Packet->ContigSize;
        do {
            CopyNow = MIN(BufLen, Needed);
            RtlCopyMemory((uchar *)Packet->Auxiliary + Offset, VirtAddr,
                          CopyNow);
            Needed -= CopyNow;
            Offset += CopyNow;
            NdisGetNextBuffer(Buffer, &Buffer);
            NdisQueryBuffer(Buffer, &VirtAddr, &BufLen);
        } while(Needed);
        Packet->ContigSize = Needed;
    }

    //
    // Point our packet's data pointer at the newly allocated
    // contiguous region.
    //
    Packet->Data = Packet->Auxiliary;
    return Packet->ContigSize;
}


#include <stdio.h>

//* ParseNumber
//
//  Parses a number, given a base and a maximum.
//  The base should be 10 or 16.
//  The maximum should be 1<<8 or 1<<16.
//
//  Returns FALSE if the parsed number exceeds the maximum.
//
int
ParseNumber(WCHAR *S, uint Base, uint MaxNumber, uint *Number)
{
    uint number;
    uint digit;

    for (number = 0;; S++) {
        if ((L'0' <= *S) && (*S <= L'9'))
            digit = *S - L'0';
        else if (Base == 16) {
            if ((L'a' <= *S) && (*S <= L'f'))
                digit = *S - L'a' + 10;
            else if ((L'A' <= *S) && (*S <= L'F'))
                digit = *S - L'A' + 10;
            else
                break;
        } else
            break;

        number = number * Base + digit;
        if (number >= MaxNumber)
            return FALSE;
    }

    *Number = number;
    return TRUE;
}

//* ParseV6Address
//
//  Parses the string S as an IPv6 address. See RFC 1884.
//  The basic string representation consists of 8 hex numbers
//  separated by colons, with a couple embellishments:
//  - a string of zero numbers (at most one) may be replaced
//  with a double-colon. Double-colons are allowed at beginning/end
//  of the string.
//  - the last 32 bits may be represented in IPv4-style dotted-octet notation.
//  For example,
//     ::
//     ::1
//     ::157.56.138.30
//     ::ffff:156.56.136.75
//     ff01::
//     ff02::2
//     0:1:2:3:4:5:6:7
//  The parser is generally pretty anal about checking the syntax.
//
//  Returns FALSE if there is a syntax error.
//
int
ParseV6Address(
    IN WCHAR *S,
    OUT WCHAR **Terminator,
    OUT IPv6Addr *Addr
    )
{
    enum { Start, InNumber, AfterDoubleColon } state = Start;
    WCHAR *number = NULL;
    int sawHex;
    int numColons = 0, numDots = 0;
    int sawDoubleColon = 0;
    int i = 0;
    WCHAR c;
    uint num;

    // There are a several difficulties here. For one, we don't know
    // when we see a double-colon how many zeroes it represents.
    // So we just remember where we saw it and insert the zeroes
    // at the end. For another, when we see the first digits
    // of a number we don't know if it is hex or decimal. So we
    // remember a pointer to the first character of the number
    // and convert it after we see the following character.

    while (c = *S) {

        switch (state) {
        case Start:
            if (c == L':') {

                // this case only handles double-colon at the beginning

                if (numDots > 0)
                    goto Finish;
                if (numColons > 0)
                    goto Finish;
                if (S[1] != L':')
                    goto Finish;

                sawDoubleColon = 1;
                numColons = 2;
                Addr->u.Word[i++] = 0; // pretend it was 0::
                S++;
                state = AfterDoubleColon;

            } else
        case AfterDoubleColon:
            if ((L'0' <= c) && (c <= L'9')) {

                sawHex = FALSE;
                number = S;
                state = InNumber;

            } else if (((L'a' <= c) && (c <= L'f')) ||
                       ((L'A' <= c) && (c <= L'F'))) {

                if (numDots > 0)
                    goto Finish;

                sawHex = TRUE;
                number = S;
                state = InNumber;

            } else
                goto Finish;
            break;

        case InNumber:
            if ((L'0' <= c) && (c <= L'9')) {

                // remain in InNumber state

            } else if (((L'a' <= c) && (c <= L'f')) ||
                       ((L'A' <= c) && (c <= L'F'))) {

                if (numDots > 0)
                    goto Finish;

                sawHex = TRUE;
                // remain in InNumber state;

            } else if (c == L':') {

                if (numDots > 0)
                    goto Finish;
                if (numColons > 6)
                    goto Finish;

                if (S[1] == L':') {

                    if (sawDoubleColon)
                        goto Finish;
                    if (numColons > 5)
                        goto Finish;

                    sawDoubleColon = numColons+1;
                    numColons += 2;
                    S++;
                    state = AfterDoubleColon;

                } else {
                    numColons++;
                    state = Start;
                }

            } else if (c == L'.') {

                if (sawHex)
                    goto Finish;
                if (numDots > 2)
                    goto Finish;
                if (numColons > 6)
                    goto Finish;
                numDots++;
                state = Start;

            } else
                goto Finish;
            break;
        }

        // If we finished a number, parse it.

        if ((state != InNumber) && (number != NULL)) {

            // Note either numDots > 0 or numColons > 0,
            // because something terminated the number.

            if (numDots == 0) {
                if (!ParseNumber(number, 16, 1<<16, &num))
                    return FALSE;
                Addr->u.Word[i++] = net_short((ushort) num);
            } else {
                if (!ParseNumber(number, 10, 1<<8, &num))
                    return FALSE;
                Addr->u.Byte[2*i + numDots-1] = (uchar) num;
            }
        }

        S++;
    }

Finish:
    // Check that we have a complete address.

    if (numDots == 0)
        ;
    else if (numDots == 3)
        numColons++;
    else
        return FALSE;

    if (sawDoubleColon)
        ;
    else if (numColons == 7)
        ;
    else
        return FALSE;

    // Parse the last number, if necessary.

    if (state == InNumber) {

        if (numDots == 0) {
            if (!ParseNumber(number, 16, 1<<16, &num))
                return FALSE;
            Addr->u.Word[i] = net_short((ushort) num);
        } else {
            if (!ParseNumber(number, 10, 1<<8, &num))
                return FALSE;
            Addr->u.Byte[2*i + numDots] = (uchar) num;
        }

    } else if (state == AfterDoubleColon) {

        Addr->u.Word[i] = 0; // pretend it was ::0

    } else
        return FALSE;

    // Insert zeroes for the double-colon, if necessary.

    if (sawDoubleColon) {

        RtlMoveMemory(&Addr->u.Word[sawDoubleColon + 8 - numColons],
                      &Addr->u.Word[sawDoubleColon],
                      (numColons - sawDoubleColon) * sizeof(ushort));
        RtlZeroMemory(&Addr->u.Word[sawDoubleColon],
                      (8 - numColons) * sizeof(ushort));
    }

    *Terminator = S;
    return TRUE;
}

//* FormatV6AddressWorker - Print an IPv6 address to a buffer.
//
//  We assume the buffer has room for the longest possible address,
//  which is 40 characters.
//
void
FormatV6AddressWorker(WCHAR *Sz, IPv6Addr *Addr)
{
    //
    // Check for IPv6-compatible and IPv4-mapped addresses
    //
    if ((Addr->u.Word[0] == 0) && (Addr->u.Word[1] == 0) &&
        (Addr->u.Word[2] == 0) && (Addr->u.Word[3] == 0) &&
        (Addr->u.Word[4] == 0) &&
        ((Addr->u.Word[5] == 0) || (Addr->u.Word[5] == 0xffff)) &&
        (Addr->u.Word[6] != 0)) {

        Sz += swprintf(Sz, L"::%s%u.%u.%u.%u",
                      Addr->u.Word[5] == 0 ? "" : "ffff:",
                      Addr->u.Byte[12], Addr->u.Byte[13],
                      Addr->u.Byte[14], Addr->u.Byte[15]);
    } else {
        int maxFirst, maxLast;
        int curFirst, curLast;
        int i;

        //
        // Find largest contiguous substring of zeroes
        // A substring is [First, Last), so it's empty if First == Last.
        //

        maxFirst = maxLast = 0;
        curFirst = curLast = 0;

        for (i = 0; i < 8; i++) {

            if (Addr->u.Word[i] == 0) {
                // Extend current substring
                curLast = i+1;

                // Check if current is now largest
                if (curLast - curFirst > maxLast - maxFirst) {

                    maxFirst = curFirst;
                    maxLast = curLast;
                }
            } else {
                // Start a new substring
                curFirst = curLast = i+1;
            }
        }

        //
        // Ignore a substring of length 1.
        //
        if (maxLast - maxFirst <= 1)
            maxFirst = maxLast = 0;

        //
        // Write colon-separated words.
        // A double-colon takes the place of the longest string of zeroes.
        // All zeroes is just "::".
        //
        for (i = 0; i < 8; i++) {
            uint word;

            // Skip over string of zeroes
            if ((maxFirst <= i) && (i < maxLast)) {

                Sz += swprintf(Sz, L"::");
                i = maxLast-1;
                continue;
            }

            // Need colon separator if not at beginning
            if ((i != 0) && (i != maxLast))
                Sz += swprintf(Sz, L":");

            // Esthetically, I don't like 3 digit words
            word = net_short(Addr->u.Word[i]);
            Sz += swprintf(Sz, word < 0x100 ? L"%x" : L"%04x", word);
        }
    }
}

#if DBG
//* FormatV6Address - Print an IPv6 address to a buffer.
//
//  Returns a static buffer containing the address.
//  This function is only useful for debug prints.
//
WCHAR *
FormatV6Address(IPv6Addr *Addr)
{
    static WCHAR Buffer[40];

    FormatV6AddressWorker(Buffer, Addr);
    return Buffer;
}
#endif // DBG
