// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Definitions for using IPv4 as a link-layer for IPv6.
//


#ifndef TUNNEL_INCLUDED
#define TUNNEL_INCLUDED 1

//
// A few IPv4 definitions that we need.
// Including the v4 header files causes errors.
// REVIEW: can we fix this?
//

#define INADDR_ANY 0

typedef unsigned long IPAddr;  // An IP address.

//
// IP Header format.
//
typedef struct IPHeader {
    uchar iph_verlen;    // Version and length.
    uchar iph_tos;       // Type of service.
    ushort iph_length;   // Total length of datagram.
    ushort iph_id;       // Identification.
    ushort iph_offset;   // Flags and fragment offset.
    uchar iph_ttl;       // Time to live.
    uchar iph_protocol;  // Protocol.
    ushort iph_xsum;     // Header checksum.
    IPAddr iph_src;      // Source address.
    IPAddr iph_dest;     // Destination address.
} IPHeader;

//
// Tunnel protocol constants.
//
#define TUNNEL_PROTOCOL    41
#define TUNNEL_DEVICE_NAME (DD_RAW_IP_DEVICE_NAME L"\\41")
#define TUNNEL_MTU         1480
#define TUNNEL_6OVER4_TTL  16

//
// Each tunnel interface (including the pseudo-interface)
// provides a TunnelContext to the IPv6 code as its link-level context.
//
// Each tunnel interface uses a separate v4 TDI address object
// for sending packets. This allows v4 attributes of the packets
// (like source address and TTL) to be controlled individually
// for each tunnel interface.
//
// The pseudo-tunnel interface does not control the v4 source address
// of the packets that it sends; the v4 stack is free to chose
// any v4 address. Note that this means a packet with a v4-compatible
// v6 source address might be sent using a v4 source address
// that is not derived from the v6 source address.
//
// The 6-over-4 and point-to-point virtual interfaces, however,
// do strictly control the v4 source addresses of their packets.
// As "real" interfaces, they participate in Neighbor Discovery
// and their link-level (v4) address is important.
//
// In contract to sending, the receive path uses a single address object
// for all tunnel interfaces. We use the TDI address object
// associated with the pseudo-interface; because it's bound to INADDR_ANY
// it receives all encapsulated v6 packets sent to the machine.
//

typedef struct TunnelContext {
    IPAddr Addr;
    Interface *IF;
    PFILE_OBJECT AOFile;
    HANDLE AOHandle;
} TunnelContext;

// Information we keep globally.
typedef struct TunnelGlobals {
    KSPIN_LOCK Lock;

    IPv6Addr DummyAddr;                 // Used internally.

    WCHAR *RegKeyNameParam;             // Where we get parameters.

    PDEVICE_OBJECT V4Device;            // v4 device object.

    TunnelContext Pseudo;               // For automatic/configured tunnels.
    NeighborCacheEntry *PseudoNCE;      // For automatic tunnels.

    // Used by TunnelReceive/TunnelReceiveComplete
    PIRP ReceiveIrp;
    TDI_CONNECTION_INFORMATION ReceiveInputInfo;
    TDI_CONNECTION_INFORMATION ReceiveOutputInfo;

    struct TunnelAddress *Addresses;
} TunnelGlobals;

extern TunnelGlobals Tunnel;

// Information we keep for each v4 address.
typedef struct TunnelAddress {
    struct TunnelAddress *Next;

    TunnelContext V6V4;      // For 6-over-4 interface.

    NetTableEntry *AutoNTE;  // v4-compatible address for automatic tunnels.
} TunnelAddress;

//
// Context information that we pass to the IPv4 stack
// when transmitting.
//
typedef struct TunnelTransmitContext {
    Interface *IF;
    PNDIS_PACKET packet;
    TA_IP_ADDRESS taIPAddress;
    TDI_CONNECTION_INFORMATION tdiConnInfo;
} TunnelTransmitContext;

//
// Tunnel functions specific to the various flavors of tunneling.
// Imported by tunnel.c.
//

extern Interface *
TunnelV6V4CreateInterface(TunnelAddress *ta);

//
// Generic tunneling support routines,
// exported from tunnel.c.
//

void
TunnelTransmit(
    void *Context,        // Pointer to tunnel interface.
    PNDIS_PACKET Packet,  // Pointer to packet to be transmitted.
    uint Offset,          // Offset from start of packet to IPv6 header.
    void *LinkAddress);   // Link-level address.

extern NTSTATUS
TunnelAddMulticastAddress(
    PFILE_OBJECT AO,
    IPAddr IfAddress,
    IPAddr MCastAddress);

extern NTSTATUS
TunnelDelMulticastAddress(
    PFILE_OBJECT AO,
    IPAddr IfAddress,
    IPAddr MCastAddress);

extern int
ParseV4Address(WCHAR *Sz, WCHAR **Terminator, IPAddr *Addr);

extern void FormatV4AddressWorker(WCHAR *Sz, IPAddr Addr);

extern WCHAR *FormatV4Address(IPAddr Addr);

#endif  // TUNNEL_INCLUDED
